"""Implementation of GeckoFitKcats class.

Support fitting of turnover numbers to proteomics data in GECKO models.

Peter Schubert, HHU Duesseldorf, CCB, September 2024
"""

import re
import pandas as pd
from collections import defaultdict

import f2xba.prefixes as pf


class GeckoFitKcats:
    """Support fitting of turnover numbers to proteomics data.

    Usage, with the dictionary measured_mpmfs of measured protein levels:

    .. code-block:: python

        eo = EcmOptimization('iJO1366_GECKO.xml)
        solution = eo.optimize()
        gfk = GeckoFitKcats(eo, 'iJO1366_predicted_kcats.xlsx')
        tot_fitted_mpmf = gfk.process_data(solution.fluxes, measured_mpmfs)
        exceeding_max_scale = gfk.update_kcats('iJO1366_fitted_kcats.xlsx', target_sat=0.5, max_scale_factor=100.0)
    """

    def __init__(self, optim, orig_kcats_fname):
        """Instantiate the GeckoFitKcats instance.

        :param optim: a reference to a EcmOptimization instance
        :type optim: :class:`EcmOptimization`
        :param str orig_kcats_fname: filename containing original turnover numbers (.xlsx)
        """
        self.optim = optim
        self.orig_kcats_fname = orig_kcats_fname

        locus2iso_rids = defaultdict(list)
        net2iso_rids = defaultdict(list)
        count = 0
        for iso_rid, data in self.optim.rdata.items():
            if len(data['gpr']) > 0:
                net_rid = data['net_rid']
                net2iso_rids[net_rid].append(iso_rid)
                count += 1
                for locus in data['mpmf_coupling']:
                    locus2iso_rids[locus].append(iso_rid)
        self.locus2iso_rids = dict(locus2iso_rids)
        self.net2iso_rids = dict(net2iso_rids)
        print(f'{count:4d} enzyme catalyzed iso-reactions '
              f'({len(net2iso_rids)} reactions, {len(locus2iso_rids)} proteins)')
        self.iso_rid_pred_mpmf = {}
        self.pred_net_rid_data = {}
        self.meas_net_rid_data = {}
        self.not_meas_rids = []
        self.fitted_kcats = {}

    @staticmethod
    def get_predicted_enzyme_mpmf(flux, protein_coupling):
        """Determine mass fraction for enzyme based on reaction flux and protein coupling

        :param flux: reaction flux through iso-reaction in mmol/gDWh
        :type flux: float >= 0.0
        :param protein_coupling:
        :type protein_coupling: dict (key: gene locus, val: coupling coefficient / float)
        :return: predicted enzyme mass fraction mg/gP
        :rtype: float >= 0.0
        """
        pred_enz_mpmf = 0.0
        for locus, mpmf_coupling in protein_coupling.items():
            pred_enz_mpmf += mpmf_coupling * abs(flux)
        return pred_enz_mpmf

    def process_data(self, fluxes, measured_mpmfs):
        """Process flux solution and proteomics data, prior to updating kcat values.

        Provided are the GECKO flux solution for a given condition and corresponding proteomics data
        in mg protein per g total protein indexed by gene locus of gene product. Processing of data is
        in preparation of fitting and updating the kcat values for the GECKO model.

        Kcat fitting can only be done for reactions that carry flux. The coupling factors between reaction flux
        and protein requirement get updated by scaling the original kcat value. The process ensures that traffic
        can be shifted between iso-reactions, when proteomics suggests another iso-enzyme to carry the flux.

        1. Based on the flux solution we identify the iso-reaction that carries the flux per active reaction.
           For all iso-reactions we determine the predicted protein cost (mpmf) based on the reaction flux.
        2. Based on the flux solution we sum up all reaction fluxes that could be routed through a protein.
        3. Based on proteomics and reaction fluxes, we identify the iso-reaction that should carry the
           flux per active reaction. The iso-reaction with the highest measured protein cost is selected.
           Measured protein costs for promiscuous enzymes are allocated as per predicted flux distribution.

        :param fluxes: reaction fluxes of GECKO solution for given condition
        :type fluxes: dict or pandas.Series, key: iso-reaction id (without `R_` prefix), val: flux in mmol/gDWh, float
        :param dict measured_mpmfs: gene loci and related protein mass fractions measured in mg protein / g total protein
        :return: tot_fitted_mpmf: protein mass fraction used for kcat fitting
        :rtype: float
        """
        # 'pred_net_rid_data' maps the iso-reaction that carries the predicted flux of a reaction
        # 'iso_rid_pred_mpmf' maps the predicted protein cost for each iso-reaction, assuming it carries the flux.
        tot_pred_mpmf = 0.0
        for net_rid, iso_rids in self.net2iso_rids.items():
            net_flux = max([fluxes[iso_rid] for iso_rid in iso_rids])
            if net_flux > 0.0:
                for iso_rid in iso_rids:
                    pred_enz_mpmf = self.get_predicted_enzyme_mpmf(net_flux, self.optim.rdata[iso_rid]['mpmf_coupling'])
                    self.iso_rid_pred_mpmf[iso_rid] = pred_enz_mpmf
                    if fluxes[iso_rid] > 0.0:
                        direction = 'rev' if re.match('.*_REV$', iso_rid) else 'fwd'
                        self.pred_net_rid_data[net_rid] = {'iso_rid': iso_rid, 'flux': net_flux, 'dir': direction}
                        tot_pred_mpmf += pred_enz_mpmf
        print(f'{len(self.pred_net_rid_data):4d} active catalyzed reactions with '
              f'total protein of {tot_pred_mpmf:.1f} mg/gP (based on GECKO simulation)')

        # 'protein_flux' maps the total flux that could be routed through a given gene product
        protein_flux = {}
        for locus, iso_rids in self.locus2iso_rids.items():
            net_rids = {self.optim.rdata[iso_rid]['net_rid'] for iso_rid in iso_rids}
            tot_flux = 0.0
            for net_rid in net_rids:
                if net_rid in self.pred_net_rid_data:
                    tot_flux += self.pred_net_rid_data[net_rid]['flux']
            if tot_flux > 0.0:
                protein_flux[locus] = tot_flux
        print(f'{len(protein_flux):4d} proteins potentially involved in active reactions')

        # `meas_net_rid_data` maps the iso-reaction that should carry the reaction flux based on proteomics
        for net_rid, data in self.pred_net_rid_data.items():
            direction = data['dir']
            max_enz_mpmf = 0.0
            selected_iso_rid = None
            for iso_rid in self.net2iso_rids[net_rid]:
                if ((direction == 'fwd' and (re.match('.*_REV$', iso_rid)) is None) or
                        (direction == 'rev' and re.match('.*_REV$', iso_rid))):
                    meas_enz_mpmf = 0.0
                    for locus in self.optim.rdata[iso_rid]['mpmf_coupling']:
                        if locus in measured_mpmfs:
                            meas_enz_mpmf += measured_mpmfs[locus] * data['flux'] / protein_flux[locus]
                    if meas_enz_mpmf > max_enz_mpmf:
                        max_enz_mpmf = meas_enz_mpmf
                        selected_iso_rid = iso_rid
            if max_enz_mpmf > 0.0:
                self.meas_net_rid_data[net_rid] = {'iso_rid': selected_iso_rid, 'meas_mpmf': max_enz_mpmf}
            else:
                self.not_meas_rids.append(net_rid)
        tot_fitted_mpmf = sum([data['meas_mpmf'] for data in self.meas_net_rid_data.values()])
        print(f'{len(self.meas_net_rid_data):4d} active catalyzed reactions using '
              f'total protein of {tot_fitted_mpmf:.1f} mg/gP (based on proteomics)')
        print(f'{len(self.not_meas_rids):4d} active catalyzed reactions have no measured proteins provided')

        return tot_fitted_mpmf

    def update_kcats(self, fitted_kcats_fname, target_sat=0.5, max_scale_factor=None, min_kcat=0.01, max_kcat=5000.0):
        """Fit turnover numbers to proteomics data and export updated turnover numbers to file.

        This requires process_data() to be executed first.

        Kcat fitting can only be done for reactions that carry flux. The coupling factors between reaction flux
        and protein requirement get updated by scaling the original kcat value. The process ensures that traffic
        can be shifted between iso-reactions, when proteomics suggests another iso-enzyme to carry the flux.

        Fitting is not applied, when max_scale_factor would be exceeded.

        In the simplest case, we have a given reaction flux and a single protein measurement. If predicted
        protein is too high, we increase the kcat value for the reaction to make the enzyme more efficient. The
        scaling factor is predicted/measured protein concentrations.

        If there are iso-reactions, we need to scale the kcat values of the iso-reactions as well, to avoid that
        any of the iso-reactions becomes 'cheaper'.

        More complex cases can appear with iso-reactions, when the model uses another iso-reaction than
        proteomics suggests. In this case we first have to increase the kcat value of the iso-reaction suggested
        by proteomics, and subsequently we adjust the scaling to the measured protein concentration.

        A further kcat scaling is applied to move the model to a given target enzyme saturation level.
        It is ensured that kcat volues fall into the min_kcat, max_kcat range

        Fitted kcat values are exported to fitted_kcats_fname

        :param str fitted_kcats_fname: filename for fitted and exported turnover numbers (.xlsx)
        :param float target_sat: average target saturation of fitted model (default: 0.5)
        :param float max_scale_factor: maximum scaling [1/factor ... factor] (default None)
        :param float min_kcat: minimal turnover number in s-1 (default: 0.01)
        :param float max_kcat: maximal turnover number in s-1 (default: 5000.0)
        :return: kcat records not scaled due to exceeding max scaling
        :rtype: dict[dict]
        """
        # load model kcat records used for flux solution
        with pd.ExcelFile(self.orig_kcats_fname) as xlsx:
            df_kcats = pd.read_excel(xlsx, sheet_name='kcats', index_col=0)
            print(f'{len(df_kcats):4d} original kcat records loaded from {self.orig_kcats_fname}')

        target_saturation_scale = self.optim.avg_enz_saturation / target_sat

        exceed_max_scale = {}
        for net_rid, meas_data in self.meas_net_rid_data.items():
            pred_iso_rid = self.pred_net_rid_data[net_rid]['iso_rid']
            meas_iso_rid = meas_data['iso_rid']

            # make proteomics suggested iso-reaction more favourable, if it is not already predicted
            if meas_iso_rid == pred_iso_rid:
                scale_favourable = 1.0
                pred_mpmf_ref = self.iso_rid_pred_mpmf[meas_iso_rid]
            else:
                scale_favourable = 1.02 * self.iso_rid_pred_mpmf[meas_iso_rid] / self.iso_rid_pred_mpmf[pred_iso_rid]
                pred_mpmf_ref = self.iso_rid_pred_mpmf[meas_iso_rid] / scale_favourable

            # fit kcat value to proteomics and rescale to selected target enzyme saturation level
            scale_factor = pred_mpmf_ref / meas_data['meas_mpmf'] * target_saturation_scale

            # do not fit when max_scale_factor is exceeded
            if max_scale_factor is not None:
                if (scale_factor > max_scale_factor) or (scale_factor < 1.0 / max_scale_factor):
                    key = meas_iso_rid if meas_iso_rid in df_kcats.index else f'{pf.R_}{meas_iso_rid}'
                    exceed_max_scale[net_rid] = {'iso_rid': meas_iso_rid, 'orig_kcat': df_kcats.at[key, 'kcat_per_s'],
                                                 'factor': scale_factor}
                    continue

            # scale kcat values of all iso reactions, that none becomes more favourable
            for iso_rid in self.net2iso_rids[net_rid]:
                if iso_rid == meas_iso_rid:
                    factor = scale_favourable * scale_factor
                else:
                    factor = 0.9 * scale_factor
                key = iso_rid if iso_rid in df_kcats.index else f'{pf.R_}{iso_rid}'
                orig_kcat = df_kcats.at[key, 'kcat_per_s']
                scaled_kcat = orig_kcat * factor
                df_kcats.at[key, 'kcat_per_s'] = scaled_kcat
                df_kcats.at[key, 'notes'] = 'auto-fitted to proteomics'
                self.fitted_kcats[iso_rid] = {'net_rid': net_rid, 'orig_kcat': orig_kcat,
                                              'scaled_kcat': scaled_kcat, 'factor': factor}

        scale_factors = [data['factor'] for data in self.fitted_kcats.values()]
        print(f'{len(self.fitted_kcats):4d} kcat records fitted. '
              f'Scale factor range [{min(scale_factors):.7f}, {max(scale_factors):.1f}]')
        print(f'{len(exceed_max_scale):4d} (net) reactions would exceed the maximum scaling factor of '
              f'{max_scale_factor} and will not be scaled')

        # rescale all other kcat values in case target saturation level is higher than avg enz sat of original model
        if target_saturation_scale < 1.0:
            all_keys = {re.sub('^R_', '', key) for key in df_kcats.index}
            balance_keys = all_keys.difference(self.fitted_kcats)
            for iso_rid in balance_keys:
                key = iso_rid if iso_rid in df_kcats.index else f'{pf.R_}{iso_rid}'
                df_kcats.at[key, 'kcat_per_s'] *= target_saturation_scale
                df_kcats.at[key, 'notes'] = 'default value adapted to lower enz saturation'
            print(f'{len(balance_keys):4d} kcat values of inactive reactions reduced due to change in saturation')

        # limit kcats to the given range
        df_kcats['kcat_per_s'].mask(df_kcats['kcat_per_s'] > max_kcat, max_kcat, inplace=True)
        df_kcats['kcat_per_s'].mask(df_kcats['kcat_per_s'] < min_kcat, min_kcat, inplace=True)

        # write fitted kcats to file
        with pd.ExcelWriter(fitted_kcats_fname) as writer:
            df_kcats.to_excel(writer, sheet_name='kcats')
            print(f'fitted kcats exported to {fitted_kcats_fname}')

        return exceed_max_scale



"""Implementation of NcbiProtein class.

Holds protein related information extracted from NCBI protein fasta export.

based on UniprotProtein
Used, in case we do not find a Uniport protein for given gene product
holds required data for creating XbaModel proteins

Peter Schubert, CCB, HHU Duesseldorf, September 2024
"""
from ..utils.calc_mw import protein_mw_from_aa_seq


class NcbiProtein:

    def __init__(self, attributes):
        self.id = attributes.get('locus_tag', '')
        self.gene_name = attributes.get('gene', '')
        self.protein_name = attributes.get('protein', '')
        self.aa_sequence = attributes['aa_sequence']
        self.length = len(attributes['aa_sequence'])
        self.mass = protein_mw_from_aa_seq(self.aa_sequence)
        self.protein_id = attributes.get('protein_id', '')
        self.genome_location = attributes.get('location', '')
        # values set to undefined, so record can stand in for UniprotProtein when creating XbaModel proteins
        self.location = ''
        self.ec_numbers = []
        self.signal_peptide = None
        self.cofactors = {}



"""Implementation of RbaOptimization class.

Support both cobrapy and gurobipy interfaces

Note: RBA is a feasibility problem. Using a bisection algorithm as implemented in RbaPy
(Bulović et al., 2019), the highest growth rate is retrived where the LP is still feasible.
RBA model includes variable bounds and stoichiometric coefficients that
are dependent on selected growth rate and medium.

Optimization problem related parameters need to be updated during optimization. This is performed
in according to SBML specifications using initial assignment procedures.

The class takes care of
- initial variable and constraints configuration
- extraction data configured in the SBML model, such as function definitions, initial assignments,
species references, XML annotations, Miriam annotations
- reconfiguration of optimization problem related paramters prior to each feasiblilty loop

Peter Schubert, HHU Duesseldorf, January 2024
"""

import re
import math
import pandas as pd
from collections import defaultdict

import sbmlxdf
import f2xba.prefixes as pf
from .rba_initial_assignments import InitialAssignments
from .optimize import Optimize, extract_net_fluxes


XML_SPECIES_NS = 'http://www.hhu.de/ccb/rba/species/ns'
XML_COMPARTMENT_NS = 'http://www.hhu.de/ccb/rba/compartment/ns'


class RbaOptimization(Optimize):
    """Optimization support for RBA models.

    Optimization support is provided via cobrapy and guropibpy interfaces for RBA models
    created by f2xba, utilizing the bisection algorithm of RbaPy (Bulović et al., 2019).
    When utilizing cobrapy, it is necessary to first load the model using SBML.
    """

    def __init__(self, fname, cobra_model=None):
        """Instantiation of RbaOptimization instance.

        :param str fname: filename of the SBML coded extended model
        :param cobra_model: reference to cobra model (default: None)
        :type cobra_model: cobra.Model
        """
        super().__init__('RBA', fname, cobra_model)

        required = {'species', 'reactions', 'unitDefs', 'funcDefs', 'initAssign', 'parameters', 'compartments'}
        missing = required.difference(set(self.m_dict))
        if len(missing) > 0:
            print(f'Missing components {missing} in SBML document!')
            raise AttributeError

        self.df_mm_data = self.get_macromolecule_data()
        self.df_enz_data = self.get_enzyme_data()
        self.enz_mm_composition = self.get_enzyme_mm_composition()
        self.initial_assignments = InitialAssignments(self)
        self.configure_rba_model_constraints()

        # for CobraPy models using cplex, switch off scaling
        if cobra_model and 'cplex' in self.model.solver.interface.__name__:
            self.model.solver.problem.parameters.read.scale.set(-1)

    def configure_rba_model_constraints(self):
        """Configure constraints related to RBA modelling

            I.e. Enzyme forward and reverse efficiencies configured at ≤ 0.0
        """
        if self.is_gpm:
            return self.gp_configure_rba_model_constraints()
        else:
            return self.cp_configure_rba_model_constraints()

    def gp_configure_rba_model_constraints(self):
        """Configure constraints related to RBA modelling for GuropiPy interface"""
        for constr in self.gpm.getConstrs():
            if re.match(pf.C_EF_, constr.ConstrName) or re.match(pf.C_ER_, constr.ConstrName):
                constr.sense = '<'
        print(f'RBA enzyme efficiency constraints configured (C_EF_xxx, C_ER_xxx) ≤ 0')

    def cp_configure_rba_model_constraints(self):
        """Configure constraints related to RBA modelling for CobraPy interface"""
        modify_constr_bounds = {pf.C_EF_: [None, 0.0], pf.C_ER_: [None, 0.0]}
        for constr in self.model.constraints:
            for cprefix, bounds in modify_constr_bounds.items():
                if re.match(cprefix, constr.name):
                    constr.lb, constr.ub = bounds
        print(f'RBA enzyme efficiency constraints configured (C_EF_xxx, C_ER_xxx) ≤ 0')

    # INITIAL DATA COLLECTION PART
    def get_macromolecule_data(self):
        mms = {}
        for sid, row in self.m_dict['species'].iterrows():
            if re.match(pf.MM_, sid):
                mm_id = re.sub(f'^{pf.MM_}', '', sid)
                refs = sbmlxdf.misc.get_miriam_refs(row.get('miriamAnnotation'), 'uniprot', 'bqbiol:is')
                uniprot = refs[0] if len(refs) > 0 else None
                xml_attrs = sbmlxdf.misc.extract_xml_attrs(row.get('xmlAnnotation'), ns=XML_SPECIES_NS)
                mw_aa = float(xml_attrs['weight_aa']) if 'weight_aa' in xml_attrs else None
                mw_kda = float(xml_attrs['weight_kDa']) if 'weight_kDa' in xml_attrs else None
                scale = float(xml_attrs['scale']) if 'scale' in xml_attrs else 1.0
                mms[mm_id] = [xml_attrs['type'], uniprot, mw_kda, mw_aa, scale]
        cols = ['type', 'uniprot', 'mw_kDa', 'mw_aa', 'scale']
        df_mms = pd.DataFrame(mms.values(), index=list(mms), columns=cols)
        df_mms.index.name = 'mm_id'
        return df_mms

    def get_enzyme_data(self):
        """Extract molecular weights of enzymes and process machines.

        From xmlAnnotation of reactions corresponding of related concentration variables

        :return: Enzyme data with name, molecular weight and scaling.
        :rtype: pandas DataFrame
        """
        enz_data = {}
        for vid, row in self.m_dict['reactions'].iterrows():
            if re.match(pf.V_EC_, vid) or re.match(pf.V_PMC_, vid):
                xml_attrs = sbmlxdf.misc.extract_xml_attrs(row.get('xmlAnnotation'), ns=XML_SPECIES_NS)
                mw_kda = float(xml_attrs['weight_kDa']) if 'weight_kDa' in xml_attrs else None
                scale = float(xml_attrs['scale']) if 'scale' in xml_attrs else 1.0
                enz_data[vid] = [mw_kda, scale]
        cols = ['mw_kDa', 'scale']
        df_enz_data = pd.DataFrame(enz_data.values(), index=list(enz_data), columns=cols)
        df_enz_data.index.name = 'id'
        return df_enz_data

    def get_enzyme_mm_composition(self):
        """Determine Enzyme stoiciometry wrt gene products from model stoichiometry.

        Note: stoichiometry wrt Enzyme and process machine concentrations
        are configured in initial model considering a given model growth rate (e.g. 1.0 h-1)
        From stoichiometric coefficients we can derive enzyme/process machine composition
        wrt to macro molecules

        :return: enzyme/process machine composition wrt macromolecules
        :rtype: dict
        """
        model_gr = self.m_dict['parameters'].loc['growth_rate'].value
        enz_mm_composition = defaultdict(dict)
        for rid, row in self.m_dict['reactions'].iterrows():
            if re.match(pf.V_EC_, rid) or re.match(pf.V_PMC_, rid):
                for sref_str in sbmlxdf.record_generator(row['reactants']):
                    params = sbmlxdf.extract_params(sref_str)
                    if re.match(pf.MM_, params['species']):
                        mm_id = re.sub(f'^{pf.MM_}', '', params['species'])
                        enz_mm_composition[rid][mm_id] = float(params['stoic']) / model_gr
        return dict(enz_mm_composition)

    def set_medium_conc(self, ex_metabolites_mmol_per_l):
        """Configure external metabolite concentrations (mmol/l).

        In the RBA, the growth medium is defined by external metabolite concentrations (mmol/L).
        These concentrations are used to determine the media-dependent transporter saturation
        levels of the uptake reactions. This determination is made using a Michaelis-Menten type
        saturation function with a default Michaelis constant (Km) of 1.0 mmol/L.
        External media concentration should be scaled to the default Km value.
        This implementation of RBA still uses uptake reactions, which will be opened for
        the specified medium.

        :param ex_metabolites_mmol_per_l: external metabolite concentrations (mmol/l)
        :type ex_metabolites_mmol_per_l: dict (key: metabolite id/str, val: concentration/float)
        """
        # open up related exchange reactions
        # self.model.medium = {self.sidx2ex_ridx.get(sidx, 0.0): 1000.0 for sidx in ex_metabolites_mmol_per_l}
        self.set_medium({self.sidx2ex_ridx.get(sidx, 0.0): 1000.0 for sidx in ex_metabolites_mmol_per_l})

        # configure RBA related medium concentrations
        medium_conc = {sid: ex_metabolites_mmol_per_l.get(re.sub(f'^{pf.M_}', '', sid), 0.0)
                       for ex_rid, sid in self.ex_rid2sid.items()}

        self.initial_assignments.set_medium(medium_conc)

    def set_growth_rate(self, growth_rate):
        """RBA optimization: set growth rate dependent model parameters

        :param float growth_rate: selected growth rate in h-1
        :return:
        """
        self.initial_assignments.set_growth_rate(growth_rate)

    def solve(self, gr_min=0.0, gr_max=1.5, bisection_tol=1e-5, max_iter=40, make_non_neg=False):
        """Solve RBA feasibility problem using bisection algorithm.

        Support both cobrapy and gurobipy interfaces.

        :param float gr_min: minimum growth rate in h-1
        :param float gr_max: maximum growth rate in h-1
        :param float bisection_tol: stop criteria based on growth rate tolerances (default: 1e-3)
        :param int max_iter:  stop criteria based on iteration number (default: 40)
        :param bool make_non_neg: converted model to non-negative variables only (default: False)
        :return: optimization solution
        :rtype: class:`Solution`
        """
        if self.is_gpm:
            return self._gp_solve(gr_min, gr_max, bisection_tol, max_iter, make_non_neg)
        else:
            return self._cp_solve(gr_min, gr_max, bisection_tol, max_iter)

    def _gp_solve(self, gr_min, gr_max, bisection_tol, max_iter, make_non_neg):
        """Solve RBA feasibility problem using bisection algorithem of RbaPy 1.0.

        Optionally we can modify the model to non-negative variables only.

        :param float gr_min: minimum growth rate to test in h-1, might have to be > 0.0 depending on targets
        :param float gr_max: maximum growth rate to test in h-1
        :param float bisection_tol: (optional, default 1e-3) stop criteria based on growth rate tolerances
        :param int max_iter: (optional, default 40) stop criteria basde on number of iterations
        :param bool make_non_neg: (optional, default False) converted model to non-negative variables only
        :return: optimization solution
        :rtype: class:`Solution`
        """
        # convert the model to non-negative variables only
        if make_non_neg:
            self.gp_model_non_negative_vars()

        # bisection algorithm to narrow in on maximum growth rate
        solution = None
        self.set_growth_rate(gr_min)
        if self.optimize().status == 'optimal':
            n_iter = 1
            while ((gr_max - gr_min) > bisection_tol) and (n_iter < max_iter):
                gr_test = gr_min + 0.5 * (gr_max - gr_min)
                self.set_growth_rate(gr_test)
                # self.gpm.reset()
                if self.optimize().status == 'optimal':
                    gr_min = gr_test
                else:
                    gr_max = gr_test
                n_iter += 1

            # collect optimiziation solution at optimum growth rate
            if (gr_max - gr_min) < bisection_tol:
                gr_opt = gr_min
                self.set_growth_rate(gr_opt)
                solution = self.optimize()

                # if solution with gr_min is no longer optimal, lower gr_opt slightly and try again
                if solution.status != 'optimal':
                    gr_opt = max(0.0, gr_opt - 0.5 * bisection_tol)
                    self.set_growth_rate(gr_opt)
                    solution = self.optimize()

                solution.fluxes = extract_net_fluxes(solution.fluxes)
                solution.objective_value = gr_opt
                solution.n_iter = n_iter
                # determine final density constraint
                col = self.gpm.getCol(self.gpm.getVarByName('V_TCD'))
                solution.density_constraints = {col.getConstr(idx).constrname: -col.getCoeff(idx)
                                                for idx in range(col.size())}
            else:
                print('no optimal growth rate')
        else:
            print(f'Problem infeasible at minimum growth of {gr_min}')
        if make_non_neg:
            self.gp_model_original()
        return solution

    def _cp_solve(self, gr_min, gr_max, bisection_tol, max_iter):
        """Solve RBA feasibility problem usin bisection algorithem of RbaPy 1.0.

        Use CobraPy model contexts, so we can support outer model contexts

        :param float gr_min: minimum growth rate to test in h-1, might have to be > 0.0 depending on targets
        :param float gr_max: maximum growth rate to test in h-1
        :param float bisection_tol: (optional, default 1e-3) stop criteria based on growth rate tolerances
        :param int max_iter: (optional, default 40) stop criteria basde on number of iterations
        :return: optimization solution
        :rtype: class:`Solution`
        """
        # first, check feasibility a minimal (or zero) growth
        # with self.model:
        self.set_growth_rate(gr_min)
        opt_value = self.model.slim_optimize()

        # bisection algorithm to narrow in on maximum growth rate
        if math.isfinite(opt_value):
            n_iter = 1
            while ((gr_max - gr_min) > bisection_tol) and (n_iter < max_iter):
                gr_test = gr_min + 0.5 * (gr_max - gr_min)
                # with self.model:
                self.set_growth_rate(gr_test)
                opt_value = self.model.slim_optimize()
                if math.isfinite(opt_value):
                    gr_min = gr_test
                else:
                    gr_max = gr_test
                n_iter += 1

            # collect optimiziation solution at optimum growth rate
            if (gr_max - gr_min) < bisection_tol:
                gr_opt = gr_min
                self.set_growth_rate(gr_opt)
                solution = self.model.optimize()

                # if solution with gr_min is no longer optimal, lower gr_opt slighlty and try again
                if solution.status != 'optimal':
                    gr_opt = max(0.0, gr_opt - 0.5 * bisection_tol)
                    solution = self.model.optimize()

                solution.objective_value = gr_opt
                solution.n_iter = n_iter
                solution.density_constraints = {met.id: -val for met, val
                                                in self.model.reactions.get_by_id(pf.V_TCD).metabolites.items()}
                return solution
            else:
                print('no optimal growth rate')
                return None
        else:
            print(f'Problem infeasible at minimum growth of {gr_min}')
            return None



"""Implementation of NcbiChromsome class.

Extract nucleotide information from NCBI using E-utils EFetch.

NCBI's Disclaimer and Copyright notice
(https://www.ncbi.nlm.nih.gov/About/disclaimer.html).

Check NCBI Entrez Programming Utilities https://www.ncbi.nlm.nih.gov/books/NBK25499

Peter Schubert, CCB, HHU Duesseldorf, January 2023
"""

import os
import re
import urllib.parse
import urllib.request

from .ncbi_feature import NcbiFeature
from .ncbi_protein import NcbiProtein


e_utils_url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'


class NcbiChromosome:

    def __init__(self, chrom_id, accession_id, ncbi_dir):
        """Initialize

        Download NCBI nucleotide information for given accession id.
        Use stored files, if found in ncbi_dir.

        Extract genome and feature information

        :param chrom_id: chromosome id
        :type chrom_id: str
        :param accession_id: 'accession.version' of genome, e.g. U00096.3 for Ecoli K-12
        :type accession_id: str
        :param ncbi_dir: directory where ncbi exports are stored
        :type ncbi_dir: str
        """
        self.chromosome_id = chrom_id
        self.accession_id = accession_id

        seq_fname = os.path.join(ncbi_dir, f'{self.chromosome_id}_{self.accession_id}_fasta.txt')
        self.fasta_cds_fname = os.path.join(ncbi_dir, f'{self.chromosome_id}_{self.accession_id}_fasta_cds_aa.txt')
        self.ft_fname = os.path.join(ncbi_dir, f'{self.chromosome_id}_{self.accession_id}_features.txt')

        # download data from NCBI, unless data exists locally
        if (not os.path.exists(seq_fname) or
                not os.path.exists(self.ft_fname) or
                not os.path.exists(self.fasta_cds_fname)):
            self.download_data('fasta', seq_fname)
            self.download_data('ft', self.ft_fname)
            self.download_data('fasta_cds_aa', self.fasta_cds_fname)
        else:
            print(f'extracting nucleotide sequence from {seq_fname}')

        # retrieve genome data from local file
        with open(seq_fname, 'r') as fh:
            self.header = fh.readline().strip()
            chrom_nt_sequence = ''.join([line.strip() for line in fh])

        # collect chromosome data
        nts = sorted(set(chrom_nt_sequence))
        self.nt_composition = {nt: chrom_nt_sequence.count(nt) for nt in nts}
        self.gc_content = (self.nt_composition['G'] + self.nt_composition['C']) / sum(self.nt_composition.values())

        # collect feature data
        self.features = self.extract_features(chrom_nt_sequence)
        self.rrnas = {locus: gp for locus, gp in self.features.items() if gp.gp_type == 'rRNA'}
        self.trnas = {locus: gp for locus, gp in self.features.items() if gp.gp_type == 'tRNA'}
        self.mrnas = {locus: gp for locus, gp in self.features.items() if gp.gp_type == 'CDS'}

        # collect protein data from fasta cds file
        self.proteins = self.extract_proteins()

    def download_data(self, rettype, fname):
        """Download data for retrival type from NCBI nucleotide database to file.

        rettype as per eFetch parameter description of NCBI Entrez Programming Utilities Help
          https://www.ncbi.nlm.nih.gov/books/NBK25499/table/chapter4.T._valid_values_of__retmode_and

        :param rettype: retrival type ('fasta', 'ft', 'fasta_cds_aa', 'fasta_cds_na')
        :type rettype: str
        :param fname: file name for fasta/feature download
        :type fname: str
        """
        ncbi_fetch_url = (e_utils_url + 'efetch.fcgi?' +
                          f'db=nuccore&id={self.accession_id}&rettype={rettype}&retmode="text"')
        with urllib.request.urlopen(ncbi_fetch_url) as response, open(fname, 'wb') as fh:
            fh.write(response.read())
            print(f'NCBI {rettype} records downloaded {self.accession_id} to: {fname}')

    def extract_features(self, chrom_nt_seq):
        """Extract features from NCBI features file.

        A Feature contains 'gene' records and
        'tRNA', 'rRNA' or 'CDS' records

        E.g. gene in E. coli U00096.3_features.txt:

            2725925	2725987	gene
                gene 	yfiS
                gene_syn	ECK4639
                locus_tag	b4783
                db_xref	ECOCYC:G0-17029
            2725925	2725987	CDS
                product	protein YfiS
                transl_table	11
                protein_id	gb|QNV50536.1||gnl|b4783|CDS2742
                transcript_id	gnl|b4783|mrna.CDS2742
                db_xref	UniProtKB/Swiss-Prot:P0DSG1

        A new Feature starts with a 'gene' start record.
            this record contains some attributes which are collected
        The gene type is contain in a subsequent reocord with start/stop/record type
            this record constains some attributes which are collected

        Note: special records can contain start/stop information related to spliced products
        Note: not all records will be collected, e.g. 'mobile_element' records will be skipped

        :param chrom_nt_seq: nucleotide sequence of chromosome
        :type chrom_nt_seq: str
        :return: dict with gene locus and related NCBI feature record
        :rtype: dict (key: gstr; val: class NCBIFeatureRecord)
        """
        skipped_types = {'repeat_region', 'mobile_element', 'rep_origin'}

        features = {}
        with open(self.ft_fname, 'r') as fh:
            fh.readline().strip()  # drop file header

            gene_data = None
            for line in fh:
                fields = line.rstrip().split('\t')

                # new record header. Records could be 'gene', 'CDS', 'rRNA', 'tRNA', ...
                if len(fields) == 3:
                    start = int(re.sub(r'\D', '', fields[0]))
                    stop = int(re.sub(r'\D', '', fields[1]))
                    record_type = fields[2]
                    # a 'gene' record starts a new feature record
                    if record_type == 'gene':
                        # collect info from previous feature and store the feature
                        if gene_data is not None:
                            locus = gene_data.collect_info(chrom_nt_seq)
                            # in case locus is not provided, at least a gene name should be provided
                            if locus is None:
                                assert gene_data.gene, "neither Locus_tag nor gene-data provided in NCBI seq record"
                                locus = gene_data.gene
                            features[re.sub(r'\W', '_', locus)] = gene_data
                        # create a new feature
                        gene_data = NcbiFeature(record_type, start, stop)

                    # if header is not a 'gene' record, open a sub-record related to previous 'gene' record
                    elif gene_data is not None and record_type not in skipped_types:
                        gene_data.add_record(record_type, start, stop)

                # collect attribute information related to the current record
                elif gene_data is not None:
                    if len(fields) == 2:
                        # this adds splicing information
                        start = int(re.sub(r'\D', '', fields[0]))
                        stop = int(re.sub(r'\D', '', fields[1]))
                        gene_data.add_region(record_type, start, stop)
                    elif len(fields) == 5:
                        # this adds attributes
                        gene_data.add_attribute(record_type, fields[3], fields[4])

        # collect final feature record at end of file
        if gene_data is not None:
            # final feature processing
            locus = gene_data.collect_info(chrom_nt_seq)
            # in case locus is not provided, at least a gene name should be provided
            if locus is None:
                assert gene_data.gene, "neither Locus_tag nor gene-data provided in NCBI seq record"
                locus = gene_data.gene
            features[re.sub(r'\W', '_', locus)] = gene_data

        return features

    def extract_proteins(self):
        """Extract protein data from NCBI fasta cds file

        Data can be used, when Uniprot Data is not available

        Example record:
        >lcl|NC_000911.1_prot_WP_010872574.1_1430 [gene=fldA] [locus_tag=SGL_RS08960]
         [protein=flavodoxin FldA] [protein_id=WP_010872574.1]
         [location=complement(1516659..1517171)] [gbkey=CDS]
        MTKIGLFYGTQTGNTETIAELIQKEMGGDSVVDMMDISQADVDDFRQYSCLIIGCPTWNVGELQSDWEGF
        YDQLDEIDFNGKKVAYFGAGDQVGYADNFQDAMGILEEKISGLGGKTVGFWPTAGYDFDESKAVKNGKFV
        GLALDEDNQPELTELRVKTWVSEIKPILQS

        :return: proteins
        :rtype: dict (key: locus, val: NcbiProtein
        """
        locus = None
        proteins = {}
        with open(self.fasta_cds_fname, 'r') as fh:
            for line in fh:
                line = line.strip()
                if re.match('>', line):
                    # store a previous record
                    if locus is not None:
                        attributes['aa_sequence'] = aa_seq
                        proteins[locus] = NcbiProtein(attributes)
                    # collect attributes from header line
                    attributes = {key: val for key, val in re.findall(r'(\w+)=([^]]*)', line)}
                    # in case locus_tag is not provided, get 'gene'
                    locus = attributes.get('locus_tag', attributes.get('gene', ''))
                    aa_seq = ''
                # collect amino acid sequence
                else:
                    aa_seq += line

        # process last record in file
        if locus is not None:
            attributes['aa_sequence'] = aa_seq
            proteins[locus] = NcbiProtein(attributes)

        return proteins



"""Implementation of NcbiFeature class.

Holds RNA related information extracted from NCBI feature and fasta files.

Peter Schubert, CCB, HHU Duesseldorf, January 2023
"""

from .ncbi_ft_record import NcbiFtRecord


map_complement = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G'}
record_types = {'gene', 'mRNA', 'CDS', 'tRNA', 'rRNA', 'ncRNA'}
attr_types = {'gene', 'locus_tag', 'old_locus_tag', 'product', 'protein_id', 'db_xref', 'EC_number', 'note'}


class NcbiFeature:
    """
    support
        - gene-tRNA, gene-rRNA, gene-mRNA-CDS  (yeast)
        - gene-tRNA, gene-mRNA, gene-CDS (ecoli)
        - length and composition of tRNA, rRNA, CDS
    """

    def __init__(self, record_type, start, stop):
        assert (record_type == 'gene')
        self.records = {record_type: NcbiFtRecord(record_type, start, stop)}
        self.cur_record = record_type
        self.forward = self.records[self.cur_record].forward
        self.length = self.records[self.cur_record].length
        self.spliced_length = 0
        self.nt_composition = {}
        self.spliced_nt_composition = {}
        self.gp_type = None
        self.locus = None
        self.gene = None
        self.product = None
        self.protein_id = None
        self.ecns = None
        self.xrefs = None
        self.old_locus = None
        self.note = None

    def add_record(self, record_type, start, stop):
        if record_type in record_types:
            if record_type not in self.records:
                self.records[record_type] = NcbiFtRecord(record_type, start, stop)
                self.cur_record = record_type
        else:
            self.cur_record = None

    def add_region(self, record_type, start, stop):
        if record_type == self.cur_record:
            self.records[self.cur_record].add_region(start, stop)

    def add_attribute(self, record_type, attr_id, attr_val):
        """Add a collected attribute/value pair to current record.

        Only specified attributes 'attr_types' will be collected

        :param record_type:
        :param attr_id:
        :param attr_val:
        """
        if record_type == self.cur_record:
            if attr_id in attr_types:
                self.records[self.cur_record].add_attribute(attr_id, attr_val)

    def collect_info(self, chrom_seq):
        """Collect gene feature information from collected records for the specific gene

        :param chrom_seq:
        :return:
        """
        self.locus = self.records['gene'].locus
        self.old_locus = self.records['gene'].old_locus
        self.length = self.records['gene'].length
        self.nt_composition = self.records['gene'].get_composition(chrom_seq)
        self.gene = getattr(self.records['gene'], 'gene', None)

        rec_ids = list(set(self.records).intersection({'CDS', 'tRNA', 'rRNA'}))
        if len(rec_ids) > 0:
            assert len(rec_ids) == 1
            self.gp_type = rec_ids[0]
            record = self.records[self.gp_type]
            self.spliced_length = record.length
            self.spliced_nt_composition = record.get_composition(chrom_seq)
            self.product = getattr(record, 'product', None)
            self.protein_id = getattr(record, 'protein_id', None)
            self.ecns = getattr(record, 'EC_number', None)
            self.xrefs = getattr(record, 'db_xref', None)
            self.note = getattr(record, 'note', None)
        return self.locus

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param attribute: attribute name
        :type attribute: str
        :param value: value to be configured
        :type value: str
        """
        setattr(self, attribute, value)



"""Implementation of FbaOptimization class.

Support functions for cobrapy and gurobipy optimization of FBA models.

Peter Schubert, HHU Duesseldorf, CCB, November 2024
"""

# gurobipy should not be a hard requirement, unless used in this context
try:
    import gurobipy as gp
except ImportError:
    gp = None
    pass

from .optimize import Optimize


class FbaOptimization(Optimize):
    """Optimization support for FBA models in context of f2xba optimization.

    Optimization support is provided via cobrapy and guropibpy interfaces for
    FBA and TFA models. When utilizing cobrapy, it is
    necessary to first load the model using SBML.
    """


    def __init__(self, fname, cobra_model=None):
        """Instantiate the FbaOptimization instance.

        :param str fname: filename of the SBML coded FBA/TFA model
        :param cobra_model: reference to cobra model (default: None)
        :type cobra_model: cobra.Model
        """
        super().__init__('FBA', fname, cobra_model)





"""Implementation of SbmlCompartment class.

based on XBAcompartment from xbanalysis package

Peter Schubert, HHU Duesseldorf, February 2023
"""

from .sbml_sbase import SbmlSBase


class SbmlCompartment(SbmlSBase):

    def __init__(self, s_compartment):
        """Instantiate SbmlCompartment instance with data from s_compartment.

        s_compartment is pandas Series with Series.name attribute set compartment id and
        several mandatory and optional attributes based on SBML specifications.

        - mandatory attributes:
            - Series.name: str - compartment id
            - 'constant': bool

        - optional attributes:
            - 'name': str - handled in parent class
            - 'sboterm': str - handled in parent class
            - 'metaid': str - handled in parent class
            - 'miriamAnnotation': str - handled in parent class
            - 'xmlAnnotation': str - handled in parent class
            - 'notes': str - handled in parent class
            - 'size': float
            - 'spatialDimension': float
            - 'units': str - unit definition id

        :param s_compartment: species data from SBML import
        :type s_compartment: pandas Series
        """
        super().__init__(s_compartment)
        self.constant = s_compartment['constant']
        if 'size' in s_compartment:
            self.size = s_compartment['size']
        if 'spatialDimension' in s_compartment:
            self.dimension = s_compartment['spatialDimension']
            if 'units' in s_compartment:
                self.units = s_compartment['units']
        else:
            self.units = s_compartment.get('units', 'dimensionless')

    def to_dict(self):
        data = super().to_dict()
        data['constant'] = self.constant

        if hasattr(self, 'size'):
            data['size'] = self.size
        if hasattr(self, 'units'):
            data['units'] = self.units
        if hasattr(self, 'dimension'):
            data['spatialDimension'] = self.dimension
        return data

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param attribute: attribute name
        :type attribute: str
        :param value: value to be configured
        :type value: str
        """
        setattr(self, attribute, value)



"""Implementation of SbmlGroup class.


Peter Schubert, HHU Duesseldorf, June 2023
"""

import sbmlxdf

from .sbml_sbase import SbmlSBase


class SbmlGroup(SbmlSBase):

    def __init__(self, s_group):
        super().__init__(s_group)
        self.id = s_group['id']
        self.kind = s_group.get('kind', 'partonomy')
        self.list_of_members = s_group.get('listMembers')
        self.id_refs = self.get_id_refs(s_group['members'])

    @staticmethod
    def get_id_refs(members):
        id_refs = set()
        if type(members) == str:
            for member_ref in sbmlxdf.record_generator(members):
                params = sbmlxdf.extract_params(member_ref)
                if 'idRef' in params:
                    id_refs.add(params['idRef'])
        return id_refs

    def to_dict(self):
        data = super().to_dict()
        data['kind'] = self.kind
        data['listMembers'] = self.list_of_members
        data['members'] = '; '.join([f'idRef={ref}' for ref in sorted(self.id_refs)])
        return data



"""Implementation of RbaTargetValue class.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""


class RbaTargetValue:

    def __init__(self, value=None, lb=None, ub=None):
        self.value = value
        self.lower_bound = lb
        self.upper_bound = ub

    @staticmethod
    def get_target_value(value_type, p_name):
        """Return a RbaTargetValue object instantiated

        value_type can contain either 'value', 'upperBound' or 'lowerBound'
        allowed is also 'upperBound, lowerBound'

        :param value_type: value type
        :type value_type: str
        :param p_name: parameter name (function / aggretate)
        :type p_name: str
        :return: Target Value object
        :rType: RbaTargetValue
        """
        value = None
        lb = None
        ub = None
        if 'value' in value_type:
            value = p_name
        else:
            if 'upperBound' in value_type:
                ub = p_name
            if 'lowerBound' in value_type:
                lb = p_name
        return RbaTargetValue(value=value, lb=lb, ub=ub)

    @staticmethod
    def from_dict(value_dict):
        target_value = RbaTargetValue()
        if 'value' in value_dict:
            target_value.value = value_dict['value']
        else:
            target_value.lower_bound = value_dict.get('lowerBound')
            target_value.upper_bound = value_dict.get('upperBound')
        return target_value

    def to_dict(self):
        target_values = {}
        if self.lower_bound is not None:
            target_values['lowerBound'] = self.lower_bound
        if self.upper_bound is not None:
            target_values['upperBound'] = self.upper_bound
        if self.value is not None:
            target_values['value'] = self.value
        return target_values

    def to_str(self):
        return ', '.join([f'{key}={val}' for key, val in self.to_dict().items()])



"""Abstract Class SbmlSBase

based on SBase from sbmlxdf package

Peter Schubert, February 2023
Computational Cell Design, HHU Duesseldorf
"""
import re
from abc import ABC, abstractmethod
from .miriam_annotation import MiriamAnnotation


class SbmlSBase(ABC):

    @abstractmethod
    def __init__(self, s_data):
        if s_data.name and type(s_data.name) is str:
            if re.search(r'\W', s_data.name) or re.match(r'\d', s_data.name):
                raise ValueError(f'Invalid component ID {s_data.name}; '
                                 'allowed chars: [a-zA-Z0-9_], ID must not start with [0-9]')
            self.id = s_data.name
        if type(s_data.get('name')) == str:
            self.name = s_data['name']
        if s_data.get('sboterm') is not None:
            self.sboterm = s_data['sboterm']
        if s_data.get('metaid') is not None:
            self.metaid = s_data['metaid']
        self.miriam_annotation = MiriamAnnotation(s_data.get('miriamAnnotation'))
        if s_data.get('xmlAnnotation') is not None:
            self.xml_annotation = s_data['xmlAnnotation']
        if s_data.get('notes') is not None:
            self.notes = s_data['notes']

    @abstractmethod
    def to_dict(self):
        data = {}
        if hasattr(self, 'id'):
            data['id'] = self.id
        if hasattr(self, 'name'):
            data['name'] = self.name
        if hasattr(self, 'sboterm'):
            data['sboterm'] = self.sboterm
        annot_str = self.miriam_annotation.get_annot_str()
        if len(annot_str) > 0:
            data['miriamAnnotation'] = annot_str
            data['metaid'] = getattr(self, 'metaid', f'meta_{self.id}')
        if hasattr(self, 'xml_annotation'):
            data['xmlAnnotation'] = self.xml_annotation
        if hasattr(self, 'notes'):
            data['notes'] = self.notes
        return data



"""Implementation of Protein data class.

holding data extracted from Kegg and Swissprot DBs

Peter Schubert, HHU Duesseldorf, February 2023
"""


class ProteinData:

    def __init__(self, s_data):
        self.id = s_data.name
        self.name = s_data['protein_name']
        if len(s_data['genes']) > 0:
            self.loci = s_data['genes'].split(' ')
        if len(s_data['ec_numbers']) > 0:
            self.ec_numbers = s_data['ec_numbers'].split(' ')
        self.mw = s_data['mw']

    def set_ec_numbers(self, ec_numbers):
        self.ec_numbers = ec_numbers



"""Implementation of RbaTargets and RbaTargetGroup classes.

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import numpy as np
import pandas as pd
import re
from collections import defaultdict
from xml.etree.ElementTree import parse, ElementTree, Element, SubElement, indent

from ..utils.rba_utils import get_target_species_from_xml, get_target_reactions_from_xml, extract_params, \
    translate_reaction_string
from .rba_target_value import RbaTargetValue


class RbaTargets:

    def __init__(self):
        self.target_groups = {}

    def import_xml(self, model_dir):
        file_name = os.path.join(model_dir, 'targets.xml')
        if os.path.exists(file_name) is True:
            root = parse(file_name).getroot()
            assert root.tag == 'RBATargets', 'RBATargets tag expected during import'
            self.target_groups = RbaTargetGroup.import_xml(root.find('listOfTargetGroups'))
        else:
            print(f'{file_name} not found!')

    def from_df(self, m_dict):
        if 'targets' in m_dict:
            self.target_groups = RbaTargetGroup.from_df(m_dict['targets'])
        else:
            print(f'targets not imported!')

    @staticmethod
    def get_precursor_metabolites(macromolecule_set, df_pmaps):
        """For DNA and RNA get precursor metabolites.

        Using information from sheet 'processing_maps', reactants are
        extracted. Precursor metabolites are extracted based on some
        heuristics (i.e. metabolites that only appear once).

        E.g. for macromolecues='rna' select 'M_atp_c, M_ctp_c, M_gtp_c, M_utp_c',
        other metabolites would be 'M_h2o_c'

        :param str macromolecule_set: macromolecule set 'dna' or 'rna'
        :param df_pmaps: processing map configuration
        :type df_pmaps: pandas.DataFrame
        :return: set of precursor metabolites and set of other metabolites
        :rtype: 2-tuple of sets of strings
        """
        reactant_srefs = defaultdict(int)
        for _, row in df_pmaps[df_pmaps['set'] == macromolecule_set].iterrows():
            reactants, _ = translate_reaction_string(row['reaction_string'])
            for sid in reactants:
                reactant_srefs[sid] += 1

        selected_sids = {sid for sid, freq in reactant_srefs.items() if freq == 1}
        other_sids = {sid for sid, freq in reactant_srefs.items() if freq > 1}
        return selected_sids, other_sids

    def get_biomass_composition(self, biomass_rid, component_type, rba_params, xba_model):
        """Retrieve biomass composition for specified component type, used for targets.

        For specified biomass reaction determine the composition with respect to a specified
        component type:
            'metabolites': returns stoichiometries of metabolites only (exluding nucleotides and amino acids)
            'dna': returns sum of stoichiometry of dxtp metabolties
            'amino_acids': returns relative concentration or tRNAs related to amino acids in biomass

        Require sheet 'processing_maps' to identify precursor metabolites for DNA and RNAs.
        Require sheet 'trna2locus' to identify tRNA mapping to amino acid metabolites, e.g.
        'M_trnaala_c' mapped to 'M_ala__L_c'

        :param str biomass_rid: biomass reaction id
        :param str component_type: 'metabolites', 'dna' or 'amino_acids'
        :param rba_params: RBA model specific parametrization
        :type rba_params: dict of pandas DataFrames
        :param xba_model: xba model based on genome scale metabolic model
        :type xba_model: class:`XbaModel`
        :return: biomass composition for selected component type
        :rtype: dict (key: sid/str, val: stoichiometry/float)
        """
        # identify precursor species ids of DNA/RNA based on 'reaction_string' in sheet 'processing_maps'
        dna_nt_sids, other_sids = self.get_precursor_metabolites('dna', rba_params['processing_maps'])
        rna_nt_sids, _ = self.get_precursor_metabolites('rna', rba_params['processing_maps'])

        # map amino acid species ids to tRNAs based on sheet 'trn2locus'
        aa2trna = {aa_sid: trna for trna, aa_sid in rba_params['trna2locus']['biomass_aa'].items()
                   if type(aa_sid) is str}

        # group reactant species ids of specified biomass reaction by dna, rna, amino_acid, metabolite
        dna_nt_stoic = {}
        rna_nt_stoic = {}
        aa_stoic = {}
        metabolite_stoic = {}
        for sid, stoic in xba_model.reactions[biomass_rid].reactants.items():
            if sid in dna_nt_sids:
                dna_nt_stoic[sid] = stoic
            elif sid in rna_nt_sids:
                rna_nt_stoic[sid] = stoic
            elif sid in aa2trna:
                aa_stoic[sid] = stoic
            # elif sid not in other_sids and sid.split('_')[-1] != xba_model.external_compartment:
            # others_sids include h2o_c, ppi_c, h_c
            elif sid not in other_sids:
                metabolite_stoic[sid] = stoic

        # depending on selected component type, determine respective biomass composition
        srefs = {}
        if component_type == 'dna':
            srefs['dna'] = sum(dna_nt_stoic.values())
        elif component_type == 'metabolites':
            srefs = metabolite_stoic
        elif component_type == 'amino_acids':
            total_aa = sum(aa_stoic.values())
            for aa, val in aa_stoic.items():
                srefs[aa2trna[aa]] = val / total_aa
        return srefs

    def from_xba(self, rba_params, xba_model, parameters):
        """Configure Target Constraints based on RBA specific parameters.

        used: rba_parameters['targets']
        functions and aggregates are added to Parameters

        concentration targets for biomass components can be generated automatically.
        As 'target' provide biomass reaction id and, comma separated the type the component
        ['metabolites', 'amino_acids', 'dna'], e.g. 'R_BIOMASS_Ec_iJO1366_WT_53p95M, metabolites'
        in 'target_constant' provide a scaling factor.

        :param rba_params: RBA model specific parametrization
        :type rba_params: dict of pandas.DataFrames
        :param xba_model: xba model based on genome scale metabolic model
        :type xba_model: class:`XbaModel`
        :param parameters: RBA model parameters
        :type parameters: class:`RbaParameters`
        """
        n_targets = 0
        for tgid in set(rba_params['targets'].index):
            rba_tg = RbaTargetGroup(tgid)
            df = rba_params['targets'].loc[[tgid]]
            for _, row in df.iterrows():
                target = row['target']
                target_type = row['target_type']
                value_type = row.get('target_value_type', 'value')
                constant = row.get('target_constant', np.nan)
                func = row.get('target_function', np.nan)
                agg = row.get('target_aggregate', np.nan)

                # special treatment of biomass related concentration targets
                # TODO: special treatment of biomass conc targets could be reworked and simplified
                if (',' in target) and (np.isfinite(constant)):
                    biomass_rid = target.split(',')[0].strip()
                    info = target.split(',', 1)[1].strip()
                    assert biomass_rid in xba_model.reactions, f'Biomass reaction {biomass_rid} not found in model'
                    srefs = {}
                    if re.match('exclude:', info):
                        exclude_sids = {item.strip() for item in (info.split(':')[1]).split(',')}
                        for sid, stoic in xba_model.reactions[biomass_rid].reactants.items():
                            if sid not in exclude_sids:
                                srefs[sid] = stoic
                    elif re.match('include:', info):
                        include_sids = {item.strip() for item in (info.split(':')[1]).split(',')}
                        for sid, stoic in xba_model.reactions[biomass_rid].reactants.items():
                            if sid in include_sids:
                                srefs[sid] = stoic
                    elif info in ['metabolites', 'dna', 'amino_acids']:
                        srefs = self.get_biomass_composition(biomass_rid, info, rba_params, xba_model)
                    # create individual targets
                    for sid, stoic in srefs.items():
                        p_name = parameters.create_parameter(f'{sid}_{target_type}_auto_target', stoic * constant,
                                                             np.nan, np.nan)
                        n_targets += 1
                        target_value = RbaTargetValue.get_target_value(value_type, p_name)
                        rba_tg.add_target(sid, target_type, target_value)
                else:
                    p_name = parameters.create_parameter(f'{target}_{target_type}_target', constant, func, agg)
                    if p_name:
                        n_targets += 1
                        target_value = RbaTargetValue.get_target_value(value_type, p_name)
                        rba_tg.add_target(target, target_type, target_value)
            self.target_groups[tgid] = rba_tg
        print(f'{n_targets:4d} targets in {len(self.target_groups)} target groups')

    def export_xml(self, model_dir):
        file_name = os.path.join(model_dir, 'targets.xml')
        root = Element('RBATargets')
        target_groups = SubElement(root, 'listOfTargetGroups')
        for item in self.target_groups.values():
            target_groups.append(item.export_xml())
        tree = ElementTree(root)
        indent(tree)
        tree.write(file_name)

    def to_df(self):
        data = []
        for tgid, tg in self.target_groups.items():
            tdict = tg.to_dict()
            for target_type in ['concentrations', 'productionFluxes', 'degradationFluxes', 'reactionFluxes']:
                for target, value in tdict[target_type].items():
                    data.append([tgid, target_type, target, value])
        df = pd.DataFrame(data, columns=['targetGroup', 'targetType', 'target', 'targetValue'])
        df.set_index('targetGroup', inplace=True)
        return {'targets': df}

    def validate(self, component_ids):
        valid = True
        missing = self.ref_molecules().difference(component_ids['species']) \
            .difference(component_ids['rna']) \
            .difference(component_ids['dna']) \
            .difference(component_ids['protein'])
        if len(missing) > 0:
            print('species/macromolecules used in targets not defined:', missing)
            valid = False

        missing = self.ref_parameters().difference(component_ids['functions']) \
            .difference(component_ids['aggregates'])
        if len(missing) > 0:
            print('function/aggregates used in targets not defined:', missing)
            valid = False

        return valid

    def ref_molecules(self):
        refs = set()
        for tg in self.target_groups.values():
            refs |= {sid for sid in tg.concentrations}
            refs |= {sid for sid in tg.production_fluxes}
            refs |= {sid for sid in tg.degradation_fluxes}
        return refs

    def ref_parameters(self):
        refs = set()
        for tg in self.target_groups.values():
            for target_type in ['concentrations', 'production_fluxes', 'degradation_fluxes', 'reaction_fluxes']:
                targets = getattr(tg, target_type)
                for target in targets.values():
                    refs |= set(target.to_dict().values())
        return refs


class RbaTargetGroup:

    def __init__(self, tgid):
        self.id = tgid
        self.concentrations = {}
        self.production_fluxes = {}
        self.degradation_fluxes = {}
        self.reaction_fluxes = {}

    def add_target(self, target_id, t_type, target_value):
        """Add a target of a specific target type to the target group.

        :param target_id: target id, e.g. species id
        :type target_id: str
        :param t_type: target type, e.g. 'concentrations' or 'productionFluxes'
        :type t_type: str
        :param target_value: target value of target
        :type target_value: RbaTargetValue
        """
        t_type2var = {'concentrations': self.concentrations,  'productionFluxes':  self.production_fluxes,
                      'degradationFluxes': self.degradation_fluxes, 'reactionFluxes': self.reaction_fluxes}
        t_type2var[t_type][target_id] = target_value

    @staticmethod
    def import_xml(target_groups):
        data = {}
        for target_group in target_groups.findall('targetGroup'):
            tgid = target_group.attrib.get('id', '')
            rba_tg = RbaTargetGroup(tgid)
            rba_tg.concentrations = get_target_species_from_xml(target_group.find('listOfConcentrations'))
            rba_tg.production_fluxes = get_target_species_from_xml(target_group.find('listOfProductionFluxes'))
            rba_tg.degradation_fluxes = get_target_species_from_xml(target_group.find('listOfDegradationFluxes'))
            rba_tg.reaction_fluxes = get_target_reactions_from_xml(target_group.find('listOfReactionFluxes'))
            data[tgid] = rba_tg
        return data

    @staticmethod
    def from_df(df):
        data = {}
        for tgid, group in df.groupby('targetGroup'):
            rba_tg = RbaTargetGroup(tgid)
            for _, row in group.iterrows():
                target_value = extract_params(row['targetValue'])
                if row['targetType'] == 'concentrations':
                    rba_tg.concentrations[row['target']] = RbaTargetValue.from_dict(target_value)
                if row['targetType'] == 'productionFluxes':
                    rba_tg.production_fluxes[row['target']] = RbaTargetValue.from_dict(target_value)
                if row['targetType'] == 'degradationFluxes':
                    rba_tg.degradation_fluxes[row['target']] = RbaTargetValue.from_dict(target_value)
                if row['targetType'] == 'reactionFluxes':
                    rba_tg.reaction_fluxes[row['target']] = RbaTargetValue.from_dict(target_value)
            data[tgid] = rba_tg
        return data

    def export_xml(self):
        target_group = Element('targetGroup', {'id': self.id})
        for target_type, tag in {'concentrations': 'listOfConcentrations',
                                 'production_fluxes': 'listOfProductionFluxes',
                                 'degradation_fluxes': 'listOfDegradationFluxes'}.items():
            targets = getattr(self, target_type)
            if len(targets) > 0:
                lo_targets = SubElement(target_group, tag)
                for target, value in targets.items():
                    attribs = value.to_dict() | {'species': target}
                    SubElement(lo_targets, 'targetSpecies', attribs)

        targets = self.reaction_fluxes
        if len(targets) > 0:
            lo_targets = SubElement(target_group, 'listOfReactionFluxes')
            for target, value in targets.items():
                attribs = value.to_dict() | {'reaction': target}
                SubElement(lo_targets, 'targetReaction', attribs)

        return target_group

    def to_dict(self):
        conc = {target: value.to_str() for target, value in self.concentrations.items()}
        prod_fluxes = {target: value.to_str() for target, value in self.production_fluxes.items()}
        degr_fluxes = {target: value.to_str() for target, value in self.degradation_fluxes.items()}
        reac_fluxes = {target: value.to_str() for target, value in self.reaction_fluxes.items()}
        return {'targetGroup': self.id, 'concentrations': conc, 'productionFluxes': prod_fluxes,
                'degradationFluxes': degr_fluxes, 'reactionFluxes': reac_fluxes}



"""Implementation of utilitys for gurobipy model inspection.

Peter Schubert, HHU Duesseldorf, July 2024
"""

import numpy as np
from collections import defaultdict

# Functions to print GurobiPy variable and linear constraint info
vbasis2name = {0: 'basic', -1: 'non-basic at lb', -2: 'non-basic at ub', '-3': 'super-basic'}
cbasis2name = {0: 'basic', -1: 'non-basic'}


def print_var_info(var):
    """Print info on a gurobipy model variable.

    Variable type, name, lower/upper bounds,
    optimal value, reduced costs and basis, if available
    Irreducible Inconsisten System (IIS) information, if available

    :param var: gurobipy model variable
    :type var: gurobipy.Var
    """
    if var:
        name = var.VarName
        x = var.x if hasattr(var, 'x') else np.nan
        rc = var.rc if hasattr(var, 'rc') else np.nan
        vbasis = vbasis2name[var.VBasis] if hasattr(var, 'VBasis') else 'no basis yet'
        iis_info = ''
        if hasattr(var, 'IISLB'):
            iis_info = f', IIS: [{var.IISLB}, {var.IISUB}] (force [{var.IISLBForce}, {var.IISUBForce}])'
        print(f'{var.VType} {name:20s} [{var.lb} - {var.ub}] -> {x}, {vbasis}, rc: {rc:.6f}{iis_info}')
    else:
        print(f'not in gurobipy model variable')


def print_constr_info(constr):
    """Print info on a gurobipy (linear) constraint.

    Constraint name,  sense, RHS
    shadow price and slack if available
    Irreducible Inconsisten System (IIS) information, if available

    :param constr: gurobipy model linear constraint
    :type constr: gurobipy.Constr
    """
    if constr:
        name = constr.ConstrName
        pi = constr.pi if hasattr(constr, 'pi') else np.nan
        slack = constr.Slack if hasattr(constr, 'Slack') else np.nan
        cbasis = cbasis2name[constr.cBasis] if hasattr(constr, 'CBasis') else 'no basis yet'
        iis_info = ''
        if hasattr(constr, 'IISConstr'):
            iis_info = f', IIS: {constr.IISConstr} (force {constr.IISConstrForce})'
        print(f'{name:20s} {constr.sense} {constr.RHS}, {cbasis}, pi: {pi:8.5f}, slack: {slack:.6f}{iis_info}')
    else:
        print(f'not in gurobipy model constraints')


def get_coefficients(gpm):
    """From Gurobi LP extract non-zero coefficients.

    :param gpm: GurobiPy model
    :type gpm: gurobipy.Model
    :return: variable ids with dict of constraint ids with coefficients
    :rtype: dict of dict
    """
    gpm.update()
    sm_coeffs = defaultdict(dict)
    for var in gpm.getVars():
        var_name = var.VarName
        col = gpm.getCol(var)
        for idx in range(col.size()):
            coeff = col.getCoeff(idx)
            constr = col.getConstr(idx)
            sm_coeffs[var_name][constr.ConstrName] = coeff
    return dict(sm_coeffs)



"""Implementation of RbaModel class used in f2xba.

Extend the XbaModel to become a resource balance constraint model (RBA).
RBA implementation is based on RbaPy (Bulović et al., 2019)

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import re
import pandas as pd
from collections import defaultdict

from .rba_macromolecules import RbaMacromolecules
from .rba_metabolism import RbaMetabolism
from .rba_parameters import RbaParameters, RbaFunction
from .rba_processes import RbaProcesses
from .rba_enzymes import RbaEnzymes
from .rba_densities import RbaDensities
from .rba_targets import RbaTargets
from .rba_medium import RbaMedium
from ..utils.calc_mw import protein_mw_from_aa_comp, rna_mw_from_nt_comp, ssdna_mw_from_dnt_comp, get_seq_composition
from ..utils.mapping_utils import valid_sbml_sid, load_parameter_file
from .initital_assignments import InitialAssignments
import f2xba.prefixes as pf

DEFAULT_GROWTH_RATE = 1.0         # h-1 (used to set initial values for growth rate dependent parameters)
DEFAULT_ENZ_SATURATION = 0.5      # dimensionless, is overwritten in RBA Parameters Excel document
DEFAULT_MICHAELIS_CONSTANT = 1.0  # mmol/l
MAX_ENZ_CONC = 10                 # mmol/gDW
MAX_MM_PROCESSING_FLUX = 10       # in µmol per gDW per h
MAX_DENSITY_SLACK = 10            # maximum denisty slack in mmol AA per gDW

XML_SPECIES_NS = 'http://www.hhu.de/ccb/rba/species/ns'
XML_REACTION_NS = 'http://www.hhu.de/ccb/rba/reaction/ns'
XML_COMPARTMENT_NS = 'http://www.hhu.de/ccb/rba/compartment/ns'

components = {'parameters', 'dna', 'rnas', 'proteins', 'metabolism',
              'processes', 'enzymes', 'densities', 'targets', 'medium'}


class RbaModel:
    """Extend a XbaModel to a resource balance constraint model.

    The implementation of RBA is based on the Python package RbaPy (Bulović et al., 2019).

    .. code-block:: python

        xba_model = XbaModel('iJO1366.xml')
        xba_model.configure('iJO1366_RBA_xba_parameters.xlsx')

        rba_model = RbaModel(xba_model)
        rba_model.configure('iJO1366_RBA_rba_parameters.xlsx')
        if rba_model.validate():
            rba_model.export('iJO1366_RBA.xml')
    """

    def __init__(self, xba_model):
        """Instantiate the RbaModel instance.

        :param xba_model: a reference to a XbaModel instance
        :type xba_model: :class:`f2xba.XbaModel`
        """
        self.model = xba_model
        self.dna = RbaMacromolecules('dna')
        self.rnas = RbaMacromolecules('rna')
        self.proteins = RbaMacromolecules('protein')
        self.metabolism = RbaMetabolism()
        self.parameters = RbaParameters()
        self.processes = RbaProcesses()
        self.densities = RbaDensities()
        self.targets = RbaTargets()
        self.enzymes = RbaEnzymes()
        self.medium = RbaMedium()
        self.cid_mappings = {}
        self.mmid2mm = {}
        self.initial_assignments = InitialAssignments(self.parameters)
        self.parameter_values = {}
        self.avg_enz_sat = None

    def get_dummy_translation_targets(self, rba_params):
        """get translation targets for dummy proteins

        extract translation targets from rba_params['compartments']
        return target data for dummy protein translation targets

        :param rba_params: rba configuration data loaded from file
        :type rba_params: dict of dataframes
        :return: Target configuration data related to dummy proteins
        :rtype: pandas.DataFrame
        """
        df_c_data = rba_params['compartments']
        dummy_proteins = self.cid_mappings['dummy_proteins']

        # create a dataframe to use as parameters to add
        df = df_c_data.loc[dummy_proteins.values()]
        df.rename(columns=lambda x: re.sub(r'^translation_', '', x), inplace=True)
        df['target_group'] = 'translation_targets'
        df['target_type'] = 'concentrations'
        df['target'] = [f'dummy_protein_{c_name}' for c_name in df.index]
        if 'target_value_type' not in df.columns:
            df['target_value_type'] = 'value'

        cols = [col for col in df.columns if re.match('target', col)]
        df = df[cols]
        df.set_index('target_group', inplace=True)
        return df

    def add_dummy_proteins(self):
        """Add dummy proteins to different compartments based on configuration.

        use average protein composition in terms of amino acids
        """
        dummy_proteins = self.cid_mappings['dummy_proteins']

        # determine average protein composition
        aa_count = defaultdict(int)
        for p in self.model.proteins.values():
            for cmp_id, count in get_seq_composition(p.aa_sequence).items():
                aa_count[cmp_id] += count
        total_count = sum(aa_count.values())
        avg_aa_len = total_count / len(self.model.proteins)
        composition = {aa: avg_aa_len * count / total_count for aa, count in aa_count.items()}

        # Amino acids with low frequency (here, selenocystein), set stoic coef to zero
        for aa in composition:
            if composition[aa] < 0.01:
                composition[aa] = 0.0

        for p_name, cid in dummy_proteins.items():
            self.proteins.add_macromolecule(p_name, cid, composition)

        # update average protein length parameter (assuming this parameter is used for dummy targets)
        f_name = 'inverse_average_protein_length'
        self.parameters.functions[f_name] = RbaFunction(f_name, f_type='constant',
                                                        f_params={'CONSTANT': 1.0 / avg_aa_len})

        print(f'{len(dummy_proteins):4d} dummy proteins added')

    @staticmethod
    def get_cid_mappings(rba_params):
        """get mapping of compartment ids.

        Depending on 'type' column in sheet 'compartments'
        - 'cytoplasm': default compartment for unassigned gene products, e.g. rRNA proteins
        - 'uptake': compartment with transporters for medium uptake, e.g. 'outer_membrane'
        - 'medium': compartment in which species for uptake are located, e.g. 'external'

        Mapping of xba model compartment ids to RBA compartment names
        Each reaction is connected to a compartment (based on its substrates)
        Transport reactions are connected to a compartment composed of a sorted concatenation
        of substrate cids, e.g. 'c-p'.

        returns model specific compartment id mappings required for RBA model. A dict with keywords
        - rcid2cid: mapping of reaction CIDs to CIDs, dict
        - uptake_rcids: reaction CIDs for transport reactions involved in medium uptake, set of str
        - medium_cid: CID where medium is located, str
        - dummy_proteins: name of dummy proteins and their CID location, dict

        e.g.  {'rcid2cid': {'c': 'c', 'e': 'e', 'p': 'p', 'c-p': 'im', 'c-e-p': 'im', 'c-e': 'im', 'e-p': 'om'},
               'uptake_rcids': {'e-p'},
               'cytoplasm_cid': 'c',
               'medium_cid': 'e',
               'dummy_proteins': {'dummy_protein_c': 'c', 'dummy_protein_e': 'e', 'dummy_protein_p': 'p',
               'dummy_protein_im': 'im', 'dummy_protein_om': 'om'}}

        :param rba_params: RBA specific parameters loaded from parameter Excel spreadsheet
        :type rba_params: dict of pandas.DataFrames
        :return: model specific compartment id mappings
        :rtype: dict
        """
        df_c_data = rba_params['compartments']

        # map compartment ids (cids) to reaction cids - mainly required for membrane compartments
        cid_mappings = {}
        cid2rcids = {}
        for cid, row in df_c_data.iterrows():
            cid2rcids[cid] = {rcid.strip() for rcid in row['reaction_cids'].split(',')}

        cid_mappings['rcid2cid'] = {}
        for cid, rcids in cid2rcids.items():
            for rcid in rcids:
                cid_mappings['rcid2cid'][rcid] = cid

        cid_mappings['uptake_rcids'] = set()
        for cid, row in df_c_data.iterrows():
            if row['keyword'] == 'uptake':
                cid_mappings['uptake_rcids'] |= cid2rcids[cid]
            elif row['keyword'] == 'medium':
                cid_mappings['medium_cid'] = cid
            elif row['keyword'] == 'cytoplasm':
                cid_mappings['cytoplasm_cid'] = cid

        # identify compartments with translation targets for dummy proteins
        cid_mappings['dummy_proteins'] = {}
        for col in df_c_data.columns:
            if re.match('translation_target', col):
                cids = df_c_data[df_c_data[col].notna()].index
                for cid in cids:
                    p_name = f'dummy_protein_{cid}'
                    if p_name not in cid_mappings['dummy_proteins']:
                        cid_mappings['dummy_proteins'][p_name] = cid
        return cid_mappings

    def prepare_xba_model(self, rba_params):
        """Prepare XBA model for conversion to RBA.

        extract data from:
            rba_params['machineries']
            rba_params['compartments']

        Add membrane compartment to the XBA model based on 'compartments' sheet
        Add gene products required for RBA machineries based on 'machineries' sheet
        Create relevant proteins in the XBA model and map cofactors
        Split reactions into (reversible) iso-reactions, each catalyzed by a single enzyme

        :param rba_params: RBA specific parameters loaded from parameter Excel spreadsheet
        :type rba_params: dict of pandas.DataFrames
        """
        # add (membrane) compartments to the xba model
        cs_config = {}
        for cid, row in rba_params['compartments'].iterrows():
            if cid not in self.model.compartments:
                cs_config[cid] = {'name': row['name']}
        self.model.add_compartments(cs_config)
        if len(cs_config) > 0:
            print(f'{len(cs_config):4d} compartment(s) added')

        # add gene products for coenzymes
        if 'coenzymes' in rba_params:
            count = self.model.add_gps(rba_params['coenzymes'].set_index('gpid', drop=False))
            print(f'{count:4d} gene product(s) added for coenzymes')
            self.model.update_gp_mappings()

        # add gene products for process machineries
        df_mach_data = rba_params['machineries']
        count = self.model.add_gps(df_mach_data[df_mach_data['set'] == 'protein'].set_index('gpid'))
        if count > 0:
            print(f'{count:4d} gene product(s) added for process machineries')
            self.model.update_gp_mappings()

        # create model proteins for newly added gene products
        count = self.model.create_proteins()
        if count > 0:
            print(f'{count:4d} protein(s) created with UniProt information')

        if self.model.cofactor_flag is True:
            count = self.model.map_protein_cofactors()
            if count > 0:
                print(f'{count:4d} cofactor(s) mapped to species ids for added protein')

        # split reactions into (reversible) isoreactions
        n_r = len(self.model.reactions)
        self.model.add_isoenzyme_reactions()
        n_isor = len(self.model.reactions)
        print(f'{n_r} reactions -> {n_isor} isoreactions, including pseudo reactions')

    def configure(self, fname):
        """Configuration with RBA configuration data.

        Accepted tables: 'general', 'trna2locus', 'coenzymes', 'compartments', 'targets',
        'functions', 'processing_maps', 'processes', 'machineries'

        :param str fname: filename of RBA configuration file (.xlsx)
        :return: success
        :rtype: bool
        """
        sheet_names = ['general', 'trna2locus', 'coenzymes', 'compartments', 'targets',
                       'functions', 'processing_maps', 'processes', 'machineries']
        required_sheets = {'general', 'trna2locus', 'compartments', 'targets',
                           'functions', 'processing_maps', 'processes', 'machineries'}
        rba_params = load_parameter_file(fname, sheet_names)
        missing_sheets = required_sheets.difference(set(rba_params.keys()))
        if len(missing_sheets) > 0:
            print(f'missing required tables {missing_sheets}')
            raise ValueError

        if self.check_rba_params_functions(rba_params) is False:
            print('ERRORs in RBA Parameter file')
            return False
        if self.check_rba_params_labels(rba_params) is False:
            print('ERRORs in RBA Parameter file')
            return False

        general_params = rba_params['general']['value'].to_dict()
        self.avg_enz_sat = general_params.get('avg_enz_sat', DEFAULT_ENZ_SATURATION)

        self.prepare_xba_model(rba_params)

        self.cid_mappings = self.get_cid_mappings(rba_params)
        self.parameters.from_xba(rba_params)
        self.densities.from_xba(rba_params, self.parameters)
        self.targets.from_xba(rba_params, self.model, self.parameters)
        df_dummy_targets = self.get_dummy_translation_targets(rba_params)
        self.targets.from_xba({'targets': df_dummy_targets}, self.model, self.parameters)
        self.metabolism.from_xba(rba_params, self.model)

        self.medium.from_xba(general_params, self.model)

        macromolecules = set(rba_params['processes']['set'].values)
        if 'dna' in macromolecules:
            self.dna.from_xba(rba_params, self.model, self.cid_mappings)
        if 'rna' in macromolecules:
            self.rnas.from_xba(rba_params, self.model, self.cid_mappings)
        self.proteins.from_xba(rba_params, self.model, self.cid_mappings)
        self.add_dummy_proteins()
        self.enzymes.from_xba(self.avg_enz_sat, self.model, self.parameters, self.cid_mappings, self.medium,
                              DEFAULT_MICHAELIS_CONSTANT)
        self.processes.from_xba(rba_params, self)

        print(f'{len(self.parameters.functions):4d} functions, {len(self.parameters.aggregates):4d} aggregates')
        print(f'>>> RBA model created')

        self.update_xba_model()
        return True

    def update_xba_model(self):
        """Based on configured RBA model implement a corresponding XBA model.

        RBA model will be exported in RBA proprietary format.
        XBA model will be exported in standardized SBML formatt.

        values of growth_rate and configured medium are used to calculate
        function values, like enzyme efficiencies, enzyme/process concentration
        requirements, targets

        Note: some xba model updates already performed in prepare_xba_model()

        Note: this is currently implemented as an add-on to the RBA modelling.
              Alternatively, we could update XBA model alongside RBA model creation.
        """
        print('update XBA model with RBA parameters')

        growth_rate = DEFAULT_GROWTH_RATE
        protect_ids = set()

        # create units required for RBA
        rba_units = {'hour': {'name': 'hour', 'units': 'kind=second, exp=1.0, scale=0, mult=3600.0'},
                     'per_h': {'name': 'per hour', 'units': 'kind=second, exp=-1.0, scale=0, mult=3600.0'},
                     'mmol_per_gDW': {'name': 'Millimoles', 'units': 'kind=mole, exp=1.0, scale=-3, mult=1.0; '
                                                                     'kind=gram, exp=-1.0, scale=0, mult=1.0'}
                     }
        for unit_id, data in rba_units.items():
            if unit_id not in self.model.unit_defs:
                data['id'] = unit_id
                self.model.add_unit_def(data)
            protect_ids.add(unit_id)

        pid = 'growth_rate'
        p_data = {'value': growth_rate, 'units': 'per_h', 'name': 'growth rate for initial parametrization'}
        self.model.add_parameter(pid, p_data)
        protect_ids.add(pid)

        # identify compartment where RBA medium is located
        xml_annot = f'ns_uri={XML_COMPARTMENT_NS}, prefix=rba, token=compartment, medium=True'
        add_annot = [['compartment', 'xml_annotation', xml_annot]]
        cols = ['component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(add_annot, columns=cols, index=[self.cid_mappings['medium_cid']])
        self.model.modify_attributes(df_modify_attrs, 'compartment')

        # mapping of macromolecule ids to RbaMacromolecule instances
        # TODO: do we actually need self.mmid2mm ?
        mm_types = {'protein': self.proteins, 'dna': self.dna, 'rna': self.rnas}
        for mm_type, mm_data in mm_types.items():
            for mm_id, mm in mm_data.macromolecules.items():
                self.mmid2mm[mm_id] = mm

        # remove biomass reactions for the model
        rids = [rid for rid, r in self.model.reactions.items() if r.kind == 'biomass']
        self.model.del_components('reactions', rids)

        self._xba_add_macromolecule_species()
        self._xba_add_rba_constraints()
        self._xba_couple_reactions()
        self._xba_add_macromolecule_processing_reactions()
        self.parameter_values = self._xba_get_parameter_values(growth_rate)
        self._xba_add_enzyme_concentration_variables(growth_rate)
        self._xba_add_pm_concentration_variables(growth_rate)
        self._xba_add_macromolecule_target_conc_variables(growth_rate)
        self._xba_add_metabolite_target_conc_variables(growth_rate)
        self._xba_add_target_density_variables()
        self._xba_add_flux_targets()
        self._xba_set_dummy_fba_objective()
        # self._xba_unblock_exchange_reactions()
        self.initial_assignments.xba_implementation(self.model)
        self.model.clean(protect_ids)

        # add additional units required for additional parameters
        unit_id = 'mmol_per_l'
        u_dict = {'id': unit_id, 'name': 'millimole per liter',
                  'units': 'kind=mole, exp=1.0, scale=-3, mult=1.0; kind=litre, exp=-1.0, scale=0, mult=1.0'}
        if unit_id not in self.model.parameters:
            self.model.add_unit_def(u_dict)

        # add some parameter values for reference (after model.clean() so they are not removed)
        add_params = {'avg_enz_sat': {'value': self.avg_enz_sat, 'name': 'average enzyme saturation level'},
                      'default_importer_km_value': {'value': DEFAULT_MICHAELIS_CONSTANT, 'name': 'importer Km value',
                                                    'units': 'mmol_per_l'}}
        for pid, p_data in add_params.items():
            self.model.add_parameter(pid, p_data)

        # modify some model attributs and create L3V2 SBML model -
        #  set fbcStrict to False, which is required to supports immediate assignments
        self.model.model_attrs['fbcStrict'] = False
        self.model.model_attrs['id'] += f'_RBA'
        if 'name' in self.model.model_attrs:
            self.model.model_attrs['name'] = f'RBA model of ' + self.model.model_attrs['name']
        self.model.sbml_container['level'] = 3
        self.model.sbml_container['version'] = 2

        print('XBA model updated with RBA configuration')

    @staticmethod
    def check_rba_params_functions(rba_params):
        """Check functions used in RBA Parameters file

        Functions used in aggregates of 'compartments', 'targets'
        of 'processes' sheets must also be defined in 'functions'

        :param rba_params:
        :return: success
        :rtype: bool
        """
        ok_flag = True

        # check defined functions
        defined_functions = set()
        if 'functions' in rba_params:
            defined_functions = set(rba_params['functions'].index)

        # collect functions used in 'compartments', 'targets' sheet
        used_functions = defaultdict(set)
        if 'compartments' in rba_params:
            for cid, row in rba_params['compartments'].iterrows():
                agg = row.get('translation_target_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'compartments: {cid}')
                agg = row.get('density_constraint_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'compartments: {cid}')

        if 'targets' in rba_params:
            for tgid, row in rba_params['targets'].iterrows():
                agg = row.get('target_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'targets: {tgid}')

        if 'processes' in rba_params:
            for pid, row in rba_params['processes'].iterrows():
                agg = row.get('capacity_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'processes: {pid}')

        # check that functions have been defined
        for fid, xids in used_functions.items():
            if fid not in defined_functions:
                print(f'{fid:35} function not defined, but used in {xids}')
                ok_flag = False

        default_fids = {'zero', 'default_spontaneous'}
        for fid in defined_functions:
            if fid not in used_functions and fid not in default_fids:
                print(f'{fid:35s} function defined, but not used')
        return ok_flag

    def check_rba_params_labels(self, rba_params):
        """Check gene labels used in RBA Parameters file.

        Labels used 'trna2locus' and 'machineries' must exist
        in model uniprot data or ncbi-data (for rnas)

        :param rba_params:
        :return: success/failure
        :rtype: bool
        """
        ok_flag = True
        if 'trna2locus' in rba_params:
            for rna_id, row in rba_params['trna2locus'].iterrows():
                if row['label'] not in self.model.ncbi_data.locus2record:
                    print(f"{row['label']:35s} not found in ncbi data")
                    ok_flag = False

        if 'machineries' in rba_params:
            for pid, row in rba_params['machineries'].iterrows():
                if row.get('set', '') == 'protein':
                    if row['label'] not in self.model.uniprot_data.locus2uid:
                        print(f"{row['label']:35s} not found in uniprot data")
                        ok_flag = False
                elif row.get('set', '') == 'rna':
                    if row['label'] not in self.model.ncbi_data.locus2record:
                        print(f"{row['label']:35s} not found in ncbi data")
                        ok_flag = False
        return ok_flag

    def _xba_add_macromolecule_species(self):
        """Add macromolecules to XBA model species (proteins, dna, rna)

        Constraint ID: 'MM_<mmid>

        Macromolecules (except of small macromolecules, i.e. single dna, mrna single components)
        are scaled at a fixed factor of 1000, i.e. µmol instead of mmol (improves LP problem) for large
        macromolecules, i.e. > 10 amino acids weight.
        The scale factor is set on the macromolecule instance.

        MIRIAM annotation and Uniport id is added
        XML annotation with equivalent amino acid weight and scale factor is added

        Note: production / degradation reaction and target concentration variables are scalled accordingly.
        """
        xml_prefix = f'ns_uri={XML_SPECIES_NS}, prefix=rba, token=macromolecule'

        mm_species = {}
        # create protein species and configure scale
        for mm_id, mm in self.proteins.macromolecules.items():
            mm.sid = pf.MM_ + valid_sbml_sid(mm_id)
            mm.scale = 1000.0 if mm.weight > 10 else 1.0
            # for indiviually modeled proteins (Uniprot data is collected)
            if mm_id in self.model.locus2uid:
                uid = self.model.locus2uid[mm_id]
                name = f'macromolecule {mm_id} ({uid})'
                miriam_annot = f'bqbiol:is, uniprot/{uid}'
                mw_kda = self.model.proteins[uid].mw / 1000.0
            # for dummy proteins, we calculate molecular weight based on aa sequence
            else:
                mw_kda = protein_mw_from_aa_comp(mm.composition) / 1000.0
                miriam_annot = None
                name = f'macromolecule {mm_id}'
            xml_annot = (f'{xml_prefix}, type={self.proteins.type}, '
                         f'weight_aa={mm.weight}, weight_kDa={mw_kda:.4f}, scale={mm.scale}')
            mm_species[mm.sid] = [name, mm.compartment, False, False, False,
                                  f'meta_{mm.sid}', miriam_annot, xml_annot]

        for mm_id, mm in self.rnas.macromolecules.items():
            mm.sid = pf.MM_ + valid_sbml_sid(mm_id)
            mm.scale = 1000.0 if mm.weight > 10 else 1.0
            mw_kda = rna_mw_from_nt_comp(mm.composition) / 1000.0
            name = f'macromolecule {mm_id}'
            xml_annot = (f'{xml_prefix}, type={self.rnas.type}, '
                         f'weight_aa={mm.weight}, weight_kDa={mw_kda:.4f}, scale={mm.scale}')
            mm_species[mm.sid] = [name, mm.compartment, False, False, False, f'meta_{mm.sid}', None, xml_annot]

        for mm_id, mm in self.dna.macromolecules.items():
            mm.sid = pf.MM_ + valid_sbml_sid(mm_id)
            mm.scale = 1000.0 if mm.weight > 10 else 1.0
            mw_kda = ssdna_mw_from_dnt_comp(mm.composition) / 1000.0
            name = f'macromolecule {mm_id}'
            xml_annot = (f'{xml_prefix}, type={self.dna.type}, '
                         f'weight_aa={mm.weight}, weight_kDa={mw_kda:.4f}, scale={mm.scale}')
            mm_species[mm.sid] = [name, mm.compartment, False, False, False, f'meta_{mm.sid}', None, xml_annot]

        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant',
                'meta_id', 'miriamAnnotation', 'xmlAnnotation']
        df_add_species = pd.DataFrame(mm_species.values(), index=list(mm_species), columns=cols)
        print(f"{len(df_add_species):4d} macromoledules to add")
        self.model.add_species(df_add_species)

    def _xba_add_rba_constraints(self):
        """Add rba constraints to XBA model.

        constrains are added as pseudo species (similare to mass balance constraints of metabolites)
        'C_PMC_<prod_id>: process machine capacities
        'C_EF_<eid>: forward enzyme capacities
        'C_ER_<eid>: reverse enzyme capacities
        'C_D_<cid>: compartment density constraints
        """
        constraints = {}
        for proc_id, proc in self.processes.processes.items():
            if 'capacity' in proc.machinery:
                constr_id = pf.C_PMC_ + valid_sbml_sid(proc_id)
                constraints[constr_id] = [f'capacity {proc_id}', self.cid_mappings['cytoplasm_cid'],
                                          False, False, False]
                proc.constr_id = constr_id

        for eid in sorted(self.enzymes.enzymes):
            e = self.enzymes.enzymes[eid]
            if len(e.mach_reactants) > 0 or len(e.mach_products) > 0:
                eidx = re.sub(r'_enzyme$', '', eid)
                if e.forward_eff != self.parameters.f_name_zero:
                    constr_id = pf.C_EF_ + valid_sbml_sid(eidx)
                    constraints[constr_id] = [f'fwd efficiency {eid}', self.cid_mappings['cytoplasm_cid'],
                                              False, False, False]
                    e.constr_id_fwd = constr_id
                if e.backward_eff != self.parameters.f_name_zero:
                    constr_id = pf.C_ER_ + valid_sbml_sid(eidx)
                    constraints[constr_id] = [f'rev efficiency {eid}', self.cid_mappings['cytoplasm_cid'],
                                              False, False, False]
                    e.constr_id_rev = constr_id

        for cid, d in self.densities.densities.items():
            constr_id = pf.C_D_ + valid_sbml_sid(cid)
            constraints[constr_id] = [f'compartment density for {cid}', cid, False, False, False]
            d.constr_id = constr_id

        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant']
        df_add_species = pd.DataFrame(constraints.values(), index=list(constraints), columns=cols)
        print(f"{len(df_add_species):4d} RBA constraints to add")
        self.model.add_species(df_add_species)

    def _xba_couple_reactions(self):
        """Couple reactions to Enzyme Efficiencies.

        i.e. respective forward enzyme efficiency added as product to enzyme catalyzed reaction
             respective reverse enzyme efficiency added as substrate to reversible enzyme catalyzed reaction

        Note: for reverse reaction enzyme coupling RBA and TRBA reactions have to be
          coupled differently to the enzyme.
            - RBA: C_ER_<ridx> coupled with -1 to reaction rid
            - TFBA: C_ER_<ridx> coupled with +1 to reaction rid_REV

        Enzyme coupling constraint:
            C_EF_<ridx>_enzyme: R_<ridx> - kcat * V_EC_<ridx> ≤ 0
            C_ER_<ridx>_enzyme: -1 R_<ridx> - kcat * V_EC_<ridx> ≤ 0
        """
        modify_attrs = []
        for eid in sorted(list(self.enzymes.enzymes)):
            e = self.enzymes.enzymes[eid]
            if len(e.mach_reactants) > 0 or len(e.mach_products) > 0:
                rid = e.reaction
                if e.forward_eff != self.parameters.f_name_zero:
                    modify_attrs.append([e.reaction, 'reaction', 'product', f'{e.constr_id_fwd}=1.0'])
                if e.backward_eff != self.parameters.f_name_zero:
                    if e.rev_reaction == rid:
                        modify_attrs.append([rid, 'reaction', 'reactant', f'{e.constr_id_rev}=1.0'])
                    else:
                        modify_attrs.append([e.rev_reaction, 'reaction', 'product', f'{e.constr_id_rev}=1.0'])

        cols = ['id', 'component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(modify_attrs, columns=cols)
        df_modify_attrs.set_index('id', inplace=True)
        print(f'{len(df_modify_attrs):4d} fwd/rev reactions to couple with enzyme efficiency constraints')
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def _xba_add_macromolecule_processing_reactions(self):
        """Create production and degradation reactions for macromolecules.

        Here, macromolecule production/degradation reactions are implement in detail, whereas these
        are lumped in the original RBA formulation, making it difficult to follow the implementation in RBApy.

        Though this will add more variables to the linear problem, the advantages of separating the
        individiual cost terms (to better understand the underlying mechanism) could improve
        understanding of the model construction.

        for specified macromolecules we add up production requirements and create a 'R_PROC_<mm_id> reaction
        for specified macromolecules we add up degradation requirements and create a 'R_DEGR_<mm_id> reaction

        processing costs are identified and added to processing constraints 'C_PMC_<pm_id>', actually
        adding respective products with calculated costs as stoic coefficients.

        reaction fluxes are in units of µmol/gDWh (note: metabolic reaction fluxes are in mmol/gDWh)
        """
        # get mapping of macromolecule to related processing reactions for production and degradation:
        mm_productions = defaultdict(list)
        mm_degradations = defaultdict(list)
        for proc_id, proc in self.processes.processes.items():
            for mm_id in proc.productions.get('inputs', []):
                mm_productions[mm_id].append(proc_id)
            for mm_id in proc.degradations.get('inputs', []):
                mm_degradations[mm_id].append(proc_id)

        # determine processing cost for productions / degradations and create processing reactions
        lb_pid_mmol = self.model.fbc_shared_pids[self.model.flux_uid][0.0]
        ub_pid_mmol = self.model.get_fbc_bnd_pid(MAX_MM_PROCESSING_FLUX, self.model.flux_uid,
                                                 'max_processing_flux_mmol')
        lb_pid_umol = self.model.get_fbc_bnd_pid(0.0, 'umol_per_gDWh', 'zero_processing_flux_umol')
        ub_pid_umol = self.model.get_fbc_bnd_pid(MAX_MM_PROCESSING_FLUX, 'umol_per_gDWh', 'max_processing_flux_umol')

        prod_reactions = {}
        degr_reactions = {}
        for mm_id in sorted(self.mmid2mm):
            mm = self.mmid2mm[mm_id]
            # seperately check production processes and degradation processes
            for proc_type, mm2maps in {'PROD': mm_productions, 'DEGR': mm_degradations}.items():
                reactants = defaultdict(float)
                products = defaultdict(float)
                if mm2maps.get(mm_id):
                    # if there is any production/degradation process create combined production/degradation fluxes
                    for pm_id in mm2maps.get(mm_id):
                        pm = self.processes.processes[pm_id]
                        pm_map_id = pm.productions['processingMap'] if proc_type == 'PROD' \
                            else pm.degradations['processingMap']
                        pm_map = self.processes.processing_maps[pm_map_id]

                        # constant processing requirements
                        for sid, s_stoic in pm_map.constant_processing.get('reactants', {}).items():
                            reactants[sid] += float(s_stoic)/mm.scale
                        for sid, s_stoic in pm_map.constant_processing.get('products', {}).items():
                            products[sid] += float(s_stoic)/mm.scale

                        # component processing requirements
                        for comp_id, comp_stoic in mm.composition.items():
                            if comp_id in pm_map.component_processings:
                                comp_proc = pm_map.component_processings[comp_id]
                                for sid, s_stoic in comp_proc.get('reactants', {}).items():
                                    reactants[sid] += comp_stoic * float(s_stoic)/mm.scale
                                for sid, s_stoic in comp_proc.get('products', {}).items():
                                    products[sid] += comp_stoic * float(s_stoic)/mm.scale

                        # processing cost, wrt. processing machine capacity, if any
                        if pm.constr_id:
                            pm_costs = 0.0
                            for comp_id, comp_stoic in mm.composition.items():
                                if comp_id in pm_map.component_processings:
                                    comp_proc = pm_map.component_processings[comp_id]
                                    pm_costs += comp_stoic * comp_proc.get('cost', 0.0)
                            products[pm.constr_id] = round(pm_costs / mm.scale, 8)

                    if proc_type == 'PROD':
                        rid = pf.R_PROD_ + valid_sbml_sid(mm_id)
                        products[mm.sid] += 1.0
                        if mm.scale == 1.0:
                            prod_reactions[rid] = [f'production reaction for {mm_id} (mmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_mmol, ub_pid_mmol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']
                        else:
                            prod_reactions[rid] = [f'production reaction for {mm_id} (µmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_umol, ub_pid_umol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']
                    else:
                        rid = pf.R_DEGR_ + valid_sbml_sid(mm_id)
                        reactants[mm.sid] += 1.0
                        if mm.scale == 1.0:
                            degr_reactions[rid] = [f'degradation reaction for {mm_id} (mmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_mmol, ub_pid_mmol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']
                        else:
                            degr_reactions[rid] = [f'degradation reaction for {mm_id} (µmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_umol, ub_pid_umol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']

        pm_reactions = prod_reactions | degr_reactions
        cols = ['name', 'reversible', 'reactants', 'products', 'scale',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pm_reactions.values(), index=list(pm_reactions), columns=cols)
        print(f'{len(df_add_rids):4d} processing reactions to add')
        self.model.add_reactions(df_add_rids)

    def _xba_get_parameter_values(self, growth_rate):
        """calculate parameter values based on given growth rate.

        :param growth_rate: growth rate in h-1 for which to determine values
        :type growth_rate: float
        :return: parameter values determined based on params
        :rtype: dict (key: function/aggregate id / str, val value / float)
        """
        medium_cid = self.cid_mappings['medium_cid']
        mid2sid = {}
        for mid in self.medium.concentrations:
            if f'{mid}_{medium_cid}' in self.model.species:
                mid2sid[mid] = f'{mid}_{medium_cid}'

        params = {mid2sid[mid]: val for mid, val in self.medium.concentrations.items() if val != 0.0}
        params['growth_rate'] = growth_rate
        values = self.parameters.get_values(params)
        print(f"{len(values):4d} parameter values calulated based on {growth_rate:.2f} h-1 growth rate and medium")
        return values

    def _couple_weights(self, srefs):
        """Calculate enzymes/process machines weight and couple to density constraints.

        :param srefs: species reference with reactants and stoic coefficient of composition
        :type srefs: dict (key: mmid / str, val: stoic / float)
        :return: density constraints affected by weight
        :rtype: dict: (key: density constraint id / str, val: stoic / float)
        """
        weights = defaultdict(float)
        for comp_id, stoic in srefs.items():
            if comp_id in self.mmid2mm:
                mm = self.mmid2mm[comp_id]
                weights[mm.compartment] += stoic * mm.weight/mm.scale

        weight_coupling = defaultdict(float)
        for cid, weight in weights.items():
            if cid in self.densities.densities:
                weight_coupling[self.densities.densities[cid].constr_id] = weight

        return dict(weight_coupling)

    def _get_dilution_srefs(self, var_id, growth_rate, composition_srefs):
        """Determine reactant or product species references for concentration variables.

        These variables are in units of µmol/gDW. Coefficients need to be scaled

        RBA macromolecue ids are converted to respective XBA species id
        Rectants/products are based on stoichiometry of machinery composition
        Rectants/products are diluted as per growth rate
        Rectants/products are scaled to match units of concentration variable (µmol_per_gDW)
        As stoichiometry is based on growth rate, data for initial assigment is collected
        as well as sprecies reference ids

        SBML Validation:
          When the value of the attribute variable in an InitialAssignment object refers to a SpeciesReference object,
          the unit of measurement associated with the InitialAssignment's math expression should be consistent with
          the unit of stoichiometry, that is, dimensionless.

        :param var_id: id of variable, e.g. 'V_EC_<eid>', required for sref id construction
        :type var_id: str
        :param growth_rate: growth rate in h-1, dilute molecule accordingly
        :type growth_rate: float
        :param composition_srefs: reactants or product species refs
        :type composition_srefs: dict (key: sid or macromolecule id, value: stoichiometry / float)
        :return: srefs with stoichiometry
        :rtype: dict
        """
        dilution_srefs = {}
        for m_sid, m_stoic in composition_srefs.items():
            if m_sid in self.mmid2mm:
                sid = self.mmid2mm[m_sid].sid
                scale = self.mmid2mm[m_sid].scale
            else:
                sid = m_sid
                scale = 1.0
            if scale == 1000.0:
                dilution_srefs[sid] = growth_rate * m_stoic
                self.initial_assignments.add_sref_ia(var_id, sid, math_var=f'growth_rate * {m_stoic} hour')
            else:
                dilution_srefs[sid] = growth_rate * m_stoic / 1000.0
                self.initial_assignments.add_sref_ia(var_id, sid,
                                                     math_var=f'growth_rate * {m_stoic/ 1000.0} hour')
        return dilution_srefs

    def _xba_add_enzyme_concentration_variables(self, growth_rate):
        """Add enzyme concentration variables.

        Enzyme concentration variables: V_EC_<ridx>
        Couple variables to macromolecule mass balances.
        Couple these variables to enzyme efficiency and process machinery
        capacity constraints.
        Couple the variables to compartment density constraints.

        Enzyme concentrations are in µmol/gDW

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        xml_prefix = f'ns_uri={XML_SPECIES_NS}, prefix=rba, token=macromolecule'

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'umol_per_gDW', 'zero_enzyme_umol_conc')
        ub_pid = self.model.get_fbc_bnd_pid(MAX_ENZ_CONC, 'umol_per_gDW', 'max_enzyme_umol_conc')
        scale = 1000.0

        conc_vars = {}
        for eid, e in self.enzymes.enzymes.items():
            if len(e.mach_reactants) > 0 or len(e.mach_products) > 0:
                eidx = re.sub(r'_enzyme$', '', valid_sbml_sid(eid))
                var_id = pf.V_EC_ + eidx
                name = f'{eid} concentration (µmol)'

                # Determine reactants/products and sref ids for enzyme concentration variable
                reactants = self._get_dilution_srefs(var_id, growth_rate, e.mach_reactants)
                products = self._get_dilution_srefs(var_id, growth_rate, e.mach_products)

                # get gene product association from proteins used in machinery
                gps = [self.model.locus2gp[locus] for locus in e.mach_reactants if locus in self.model.locus2gp]
                gpa = 'assoc=(' + ' and '.join(gps) + ')' if len(gps) > 0 else None

                # enzyme efficiency constraints coupling
                if e.constr_id_fwd:
                    reactants[e.constr_id_fwd] = self.parameter_values[e.forward_eff] / scale
                    self.initial_assignments.add_sref_ia(var_id, e.constr_id_fwd, rba_pid=e.forward_eff,
                                                         math_const=f'{1.0/scale} dimensionless')
                if e.constr_id_rev:
                    reactants[e.constr_id_rev] = self.parameter_values[e.backward_eff] / scale
                    self.initial_assignments.add_sref_ia(var_id, e.constr_id_rev, rba_pid=e.backward_eff,
                                                         math_const=f'{1.0/scale} dimensionless')

                # couple weights of enzyme composition to density constraints
                products.update(self._couple_weights(e.mach_reactants))

                # determine enzyme molecular weight
                enz_mw_kda = 0.0
                for comp_id, stoic in e.mach_reactants.items():
                    if comp_id in self.proteins.macromolecules:
                        if comp_id in self.model.locus2uid:
                            uid = self.model.locus2uid[comp_id]
                            mw_kda = self.model.proteins[uid].mw / 1000.0
                            enz_mw_kda += stoic * mw_kda
                xml_annot = f'{xml_prefix}, weight_kDa={enz_mw_kda:.4f}, scale={scale}'

                conc_vars[var_id] = [name, False, reactants, products, scale, gpa, lb_pid, ub_pid,
                                     'RBA_enz_conc', xml_annot]

        cols = ['name', 'reversible', 'reactants', 'products', 'scale', 'fbcGeneProdAssoc',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'xmlAnnotation']
        df_add_rids = pd.DataFrame(conc_vars.values(), index=list(conc_vars), columns=cols)
        print(f'{len(df_add_rids):4d} enzyme concentration variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_pm_concentration_variables(self, growth_rate):
        """Add process machine concentration variables.

        see _xba_add_enzyme_concentration_variables
        Process machinery concentration variables: V_PMC_<pm_id>
        Couple variables to macromolecule mass balances.
        Couple these variables to enzyme efficiency and process machinery
        capacity constraints.
        Couple the variables to compartment density constraints.

        Process machinery concentrations are in µmol/gDW

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        xml_prefix = f'ns_uri={XML_SPECIES_NS}, prefix=rba, token=macromolecule'

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'umol_per_gDW', 'zero_enzyme_umol_conc')
        ub_pid = self.model.get_fbc_bnd_pid(MAX_ENZ_CONC, 'umol_per_gDW', 'max_enzyme_umol_conc')
        scale = 1000.0

        conc_vars = {}
        for pm_id, pm in self.processes.processes.items():
            if 'capacity' in pm.machinery:
                var_id = pf.V_PMC_ + valid_sbml_sid(pm_id)
                name = f'{pm_id} concentration'

                # Determine reactants/products and sref ids for PM concentration variable
                reactants = self._get_dilution_srefs(var_id, growth_rate, pm.machinery['reactants'])
                products = self._get_dilution_srefs(var_id, growth_rate, pm.machinery['products'])

                # get gene product association
                gps = [self.model.locus2gp[locus] for locus in pm.machinery['reactants']
                       if locus in self.model.locus2gp]
                gpa = 'assoc=(' + ' and '.join(gps) + ')' if len(gps) > 0 else None

                # process machinery capacity coupling
                reactants[pm.constr_id] = self.parameter_values[pm.machinery['capacity'].value] / scale
                self.initial_assignments.add_sref_ia(var_id, pm.constr_id, rba_pid=pm.machinery['capacity'].value,
                                                     math_const=f'{1.0 / scale} dimensionless')

                # couple weights of process machine composition to density constraints
                products.update(self._couple_weights(pm.machinery['reactants']))

                # determine molecular weight of process machine, consisting of proteins and/or RNA
                pm_mw_kda = 0.0
                for comp_id, stoic in pm.machinery['reactants'].items():
                    if comp_id in self.proteins.macromolecules:
                        if comp_id in self.model.locus2uid:
                            uid = self.model.locus2uid[comp_id]
                            mw_kda = self.model.proteins[uid].mw / 1000.0
                            pm_mw_kda += stoic * mw_kda
                    elif comp_id in self.rnas.macromolecules:
                        mm = self.rnas.macromolecules[comp_id]
                        mw_kda = rna_mw_from_nt_comp(mm.composition) / 1000.0
                        pm_mw_kda += stoic * mw_kda
                xml_annot = f'{xml_prefix}, weight_kDa={pm_mw_kda:.4f}, scale={scale}'

                conc_vars[var_id] = [name, False, reactants, products, scale, gpa, lb_pid, ub_pid,
                                     'RBA_pm_conc', xml_annot]

        cols = ['name', 'reversible', 'reactants', 'products', 'scale', 'fbcGeneProdAssoc',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'xmlAnnotation']
        df_add_rids = pd.DataFrame(conc_vars.values(), index=list(conc_vars), columns=cols)
        print(f'{len(df_add_rids):4d} process machine concentration variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_macromolecule_target_conc_variables(self, growth_rate):
        """Add variables for macromolecule target concentrations to XBA model.        :

        V_TMMC_<mid>: For each target concentration of a macromolecule a separate variable is introduced.
        Units: µmol/gDW or mmol/gDW
        This variable is fixed (lower/upper bound) to target concentration in respective unit (this can be
        a function of growth rate). Note: targets concentrations are defined in mmol/gDW and need to be rescaled
        Reactant coefficient for macromolecule is 1 times growth rate (i.e. macromolecular dilution),
        Product coefficient / Weight coupling is macromolecular weight divided by scale (weight in mmol AA/gDW)
        Variable bounds need to be updated during bisection optimization.

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        conc_targets = {}
        for tg_id, tg in self.targets.target_groups.items():
            for tid, t in tg.concentrations.items():
                if tid in self.mmid2mm:
                    var_id = pf.V_TMMC_ + valid_sbml_sid(tid)
                    mm = self.mmid2mm[tid]

                    # macromolecule dilution with growth rate
                    reactants = {mm.sid: growth_rate}
                    self.initial_assignments.add_sref_ia(var_id, mm.sid, math_var='growth_rate * 1 hour')

                    # add compartment density constraint
                    cid = mm.compartment
                    products = ({self.densities.densities[cid].constr_id: mm.weight / mm.scale}
                                if cid in self.densities.densities else {})

                    # get gene product association
                    gpid = self.model.locus2gp.get(tid)
                    gpa = f'assoc={gpid}' if gpid else None

                    # target concentration implemented as fixed variable bound
                    units_id = 'mmol_per_gDW' if mm.scale == 1.0 else 'umol_per_gDW'
                    var_bnd_val = self.parameter_values[t.value] * mm.scale
                    var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, units_id, f'target_conc_{tid}', reuse=False)
                    conc_targets[var_id] = [f'{tid} concentration target', False, reactants, products,
                                            mm.scale, gpa, var_bnd_pid, var_bnd_pid, 'RBA_mm_conc']
                    self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.value,
                                                            math_const=f'{mm.scale} {units_id}')

        cols = ['name', 'reversible', 'reactants', 'products', 'scale', 'fbcGeneProdAssoc',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind']
        df_add_rids = pd.DataFrame(conc_targets.values(), index=list(conc_targets), columns=cols)
        print(f'{len(df_add_rids):4d} macromolecule concentration target variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_metabolite_target_conc_variables(self, growth_rate):
        """Add variables for metabolite (small molecules) target concentrations to XBA model.        :

        V_TSMC: Target concentrations for small molecules are collected in a single variable
        Unit: µmol/gDW
        The variable is fixed at the growth rate (lower/upper bound)
        Reactant coefficients are the target concentrations (these can be functions of growth rate by itself)
        Growth rate dependent coefficients and variable bounds need to be updated during bisection optimization.
           NOTE: THIS will be changed to target = 1 and dilution terms for molecules

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        # target concentration variables are constants, i.e. fixed at value 1
        var_id = pf.V_TSMC
        one_umol_pid = self.model.get_fbc_bnd_pid(1.0e-3, 'mmol_per_gDW', 'one_umol_per_gDW', reuse=False)
        scale = 1000.0

        conc_targets = {}
        reactants_tsmc = {}
        for tg_id, tg in self.targets.target_groups.items():
            for sid, t in tg.concentrations.items():
                sbml_sid = valid_sbml_sid(sid)
                if sid not in self.mmid2mm:
                    assert sid in self.model.species, f'targets: metabolite {sid} is not included in model species'
                    reactants_tsmc[sbml_sid] = growth_rate * self.parameter_values[t.value] * scale
                    self.initial_assignments.add_sref_ia(var_id, sbml_sid, rba_pid=t.value,
                                                         math_var=f'growth_rate * {scale} hour')
        conc_targets[var_id] = [f'small molecules concentration targets (mmol)', False, reactants_tsmc, {},
                                scale, one_umol_pid, one_umol_pid, 'RBA_sm_conc']

        cols = ['name', 'reversible', 'reactants', 'products', 'scale',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind']
        df_add_rids = pd.DataFrame(conc_targets.values(), index=list(conc_targets), columns=cols)
        print(f'{len(df_add_rids):4d} metabolites concentration target variable to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_target_density_variables(self):
        """Add variables for target compartment density and slack variable to XBA model.        :

        V_TCD: Target density concentrations
        Units: mmol AA/gDW
        The variable is fixed at 1.
        Reactant coefficients are the target densities for the various compartments
        Target concentrations that depend on growth rate need to be updated during bisection aptimization

        V_SLACK_<cid>: positive slack on compartment density constraints
        Units: mmol AA/gDW
        Variable is bounded by upper value of MAX_DENSITY_SLACK (here 10 mmol AA/gDW)
        Implemented, so we can report on maximum compartment capacity utilization.
        """
        # add compartment density targets
        var_id = pf.V_TCD
        one_mmol_pid = self.model.get_fbc_bnd_pid(1.0, 'mmol_per_gDW', 'one_mmol_aa_per_gDW', reuse=False)
        density_vars = {}
        reactants = {}
        reactants_vars = {}
        for cid, d in self.densities.densities.items():
            reactants[d.constr_id] = self.parameter_values[d.target_value.upper_bound]
            self.initial_assignments.add_sref_ia(var_id, d.constr_id, rba_pid=d.target_value.upper_bound)

            fid = d.target_value.upper_bound
            if not (fid in self.parameters.functions and self.parameters.functions[fid].type == 'constant'):
                reactants_vars[d.constr_id] = fid
        density_vars[var_id] = [f'max compartment density target (mmol AA)', False, reactants, {},
                                one_mmol_pid, one_mmol_pid, 'RBA_max_density']

        # add slack variables for density constraints
        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'mmol_per_gDW', 'density_slack_min', reuse=False)
        ub_pid = self.model.get_fbc_bnd_pid(MAX_DENSITY_SLACK, 'mmol_per_gDW', 'density_slack_max', reuse=False)
        for cid, d in self.densities.densities.items():
            products = {d.constr_id: 1.0}
            density_vars[pf.V_SLACK_ + cid] = [f'Positive Slack on {cid} density', False, {}, products,
                                               lb_pid, ub_pid, 'slack_variable']

        cols = ['name', 'reversible', 'reactants', 'products',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind']
        df_add_rids = pd.DataFrame(density_vars.values(), index=list(density_vars), columns=cols)
        print(f'{len(df_add_rids):4d} density target and slack variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_flux_targets(self):
        """Add RBA reaction/production/degradation flux targets to XBA model.

        Such targets are implemented as reaction/variable bounds (fbc_bounds)

        reactionFlux targets are applied to model reactions
        productionFLux and degradatationFlux are appliedd on macromolecule production/degradation reactions.
        Note: a productionFlux target is implemented as a lower bound.

        reaction/variable bound parameters are created with 'correct' units.

        We also collect the flux bounds depending on growth rate. So they can be applied during optimization.
        """
        non_const_fids = defaultdict(dict)

        modify_attrs = []
        for tg_id, tg in self.targets.target_groups.items():

            flux_targets = {'': tg.reaction_fluxes, 'R_PROD_': tg.production_fluxes, 'R_DEGR_': tg.degradation_fluxes}
            for rid_prefix, targets in flux_targets.items():
                for tid, t in targets.items():
                    rid = f'{rid_prefix}{tid}'
                    r = self.model.reactions[rid]
                    scale = r.scale
                    ia_units = 'umol_per_gDWh' if scale == 1000.0 else 'mmol_per_gDW_per_hr'
                    unit_id = self.model.parameters[r.fbc_lower_bound].units

                    if t.value:
                        var_bnd_val = self.parameter_values[t.value] * scale
                        var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, unit_id, f'{rid}_rba_bnd', reuse=False)
                        modify_attrs.append([rid, 'reaction', 'fbc_lower_bound',    var_bnd_pid])
                        self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.value,
                                                                math_const=f'{scale} {ia_units}')
                        fid = t.value
                        if self.parameters.functions[fid].type != 'constant':
                            non_const_fids[rid]['lb'] = fid

                        if rid_prefix != 'R_PROD_':
                            modify_attrs.append([rid, 'reaction', 'fbc_upper_bound', var_bnd_pid])
                            if self.parameters.functions[fid].type != 'constant':
                                non_const_fids[rid]['ub'] = fid
                    else:
                        if t.lower_bound:
                            var_bnd_val = self.parameter_values[t.lower_bound] * scale
                            var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, unit_id, f'{rid}_rba_lb', reuse=False)
                            modify_attrs.append([rid, 'reaction', 'fbc_lower_bound', var_bnd_pid])
                            self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.lower_bound,
                                                                    math_const=f'{scale} {ia_units}')
                            fid = t.lower_bound
                            if self.parameters.functions[fid].type != 'constant':
                                non_const_fids[rid]['lb'] = fid

                        if t.upper_bound:
                            var_bnd_val = self.parameter_values[t.upper_bound] * scale
                            var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, unit_id, f'{rid}_rba_ub', reuse=False)
                            modify_attrs.append([rid, 'reaction', 'fbc_upper_bound', var_bnd_pid])
                            self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.upper_bound,
                                                                    math_const=f'{scale} {ia_units}')
                            fid = t.upper_bound
                            if self.parameters.functions[fid].type != 'constant':
                                non_const_fids[rid]['ub'] = fid

        df_modify_attrs = pd.DataFrame(modify_attrs, columns=['id', 'component', 'attribute', 'value'])
        df_modify_attrs.set_index('id', inplace=True)
        print(f"{len(df_modify_attrs):4d} Flux/variable bounds need to be updated.")
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def _xba_set_dummy_fba_objective(self):
        """Set FBA objective for RBA optimization.

        RBA is an LP feasibility problem where, where coefficients and bounds are reconfigured
        depending on a selected growth rate.
        The optmization algorithm of RBA selects for the highest growth rate, which still
        results in a feasible problem. FBA objective is not relevenat, though
        in the contexts of FBA optimization (here, just feasibility checking) an (arbitraty)
        FBA objective is defined, so problem can be solved using LP solvers
        """
        # remove any existing objectives
        obj_ids = [obj_id for obj_id in self.model.objectives]
        self.model.del_components('objectives', obj_ids)

        # Configure a dummy FBA objective:
        objectives_config = {'rba_obj': {'type': 'maximize', 'active': True, 'coefficients': {pf.V_TSMC: 1.0}}}
        self.model.add_objectives(objectives_config)
        print(f'Dummy FBA objective configured: maximize {pf.V_TSMC}')

    def _xba_unblock_exchange_reactions_old(self):
        """Unblock import of metabolites.

        In RBA formulation, nutrient environment is provided as metabolite concentrations (medium).
        Medium uptake in RBA is controlled via Michaelis Mentent saturation terms in Importer efficiencies.
        I.e. Importer efficiency is set to zero (import reaction is blocked) if a specific metabolite is
        not part of the medium.

        Original RBA formulation (RBApy) removes external metabolites from species and
        from reaction reactants in the linear problem formulation.

        In our formulation we keep the external metabolites in the problem formulation and also in
        the reactants of reactions. Import is also controlled with Michaelis Menten saturation terms
        in importer efficiencies. Exchange reactions need to be unblocked to ensure free flux of metabolites.
        """
        min_flux_val = min(self.model.fbc_flux_range)
        lb_pid = self.model.get_fbc_bnd_pid(min_flux_val, self.model.flux_uid, f'lower_exchange_flux_bnd')

        modify_attrs = []
        for rid, r in self.model.reactions.items():
            if r.kind == 'exchange':
                modify_attrs.append([rid, 'reaction', 'fbc_lower_bound', lb_pid])
        df_modify_attrs = pd.DataFrame(modify_attrs, columns=['id', 'component', 'attribute', 'value'])
        df_modify_attrs.set_index('id', inplace=True)
        print(f"{len(df_modify_attrs):4d} Exchange reactions to unblocked.")
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def to_df(self):
        m_dict = {}
        for component in components:
            m_dict |= getattr(self, component).to_df()
        return m_dict

    def to_excel(self, fname):
        """Export RBA model in RBApy format

        :param str fname: file name of spreadsheet
        :return: success (always True for now)
        :rtype: bool
        """
        m_dict = self.to_df()

        # add target value information strings
        for idx, row in m_dict['enzymes'].iterrows():
            m_dict['enzymes'].at[idx, 'fwd_eff_info'] = self.parameters.get_value_info(row['forwardEfficiency'])
            m_dict['enzymes'].at[idx, 'bwd_eff_info'] = self.parameters.get_value_info(row['backwardEfficiency'])
        for idx, row in m_dict['processes'].iterrows():
            if '=' in row['machineryCapacity']:
                first_value = row['machineryCapacity'].split('=')[1]
                m_dict['processes'].at[idx, 'capacity_info'] = self.parameters.get_value_info(first_value)
        for idx, row in m_dict['densities'].iterrows():
            first_value = row['targetValue'].split('=')[1]
            m_dict['densities'].at[idx, 'value_info'] = self.parameters.get_value_info(first_value)
        # we temporarily create a unique index for table update
        m_dict['targets'].reset_index(inplace=True)
        for idx, row in m_dict['targets'].iterrows():
            first_value = row['targetValue'].split('=')[1]
            m_dict['targets'].at[idx, 'value_info'] = self.parameters.get_value_info(first_value)
        m_dict['targets'].set_index('targetGroup', inplace=True)

        with pd.ExcelWriter(fname) as writer:
            for name, df in m_dict.items():
                df.to_excel(writer, sheet_name=name)
        print(f'RBA model exported to {fname}')
        return True

    def check_unused(self):
        molecules = (set(self.metabolism.species) | set(self.dna.macromolecules) |
                     set(self.rnas.macromolecules) | set(self.proteins.macromolecules))
        parameters = set(self.parameters.functions) | set(self.parameters.aggregates)

        ref_parameters = set()
        ref_parameters |= self.processes.ref_parameters()
        ref_parameters |= self.densities.ref_parameters()
        ref_parameters |= self.targets.ref_parameters()
        ref_parameters |= self.enzymes.ref_parameters()
        ref_parameters |= self.parameters.ref_functions(ref_parameters)

        ref_molecules = set()
        ref_molecules |= self.metabolism.ref_molecules()
        ref_molecules |= self.processes.ref_molecules()
        ref_molecules |= self.enzymes.ref_molecules()
        ref_molecules |= self.targets.ref_molecules()

        unused_parameters = parameters.difference(ref_parameters)
        unused_molecules = molecules.difference(ref_molecules)
        unused = 0
        if len(unused_parameters) > 0:
            print(f'{len(unused_parameters)} unused parameters:', unused_parameters)
            unused += len(unused_parameters)
        if len(unused_molecules) > 0:
            print(f'{len(unused_molecules)} unused molecules:', unused_molecules)
            unused += len(unused_molecules)
        if unused == 0:
            print('no unused parameters/molecules')

    def validate(self, validate_sbml=True):
        """Validate RBA configuration and compliance to SBML standards.

        Check completeness of RBA configuration and check compliance to SBML standards.

        :param bool validate_sbml: validate SBML compliance (default: True)
        :return: success
        :rtype: bool
        """
        component_ids = {'species': set(self.metabolism.species),
                         'dna': set(self.dna.macromolecules),
                         'rna': set(self.rnas.macromolecules),
                         'protein': set(self.proteins.macromolecules),
                         'functions': set(self.parameters.functions),
                         'aggregates': set(self.parameters.aggregates)}
        valid = True
        _components = {'parameters', 'metabolism', 'processes', 'enzymes', 'densities', 'targets'}
        for component in _components:
            valid = valid and getattr(self, component).validate(component_ids)
        print(f'RBA model valid status: {valid}')

        if valid is True and validate_sbml is True:
            print(f'checking SBML compliance ...')
            return self.model.validate()
        else:
            return valid

    def export(self, fname, export_format='SBML'):
        """Export RbaModel to SBML coded file, to RBA formated directory or in tabular formats.

        :param str fname: filename (with extension '.xml' or '.xlsx') or a directory name
        :param str export_format: export format: 'SBML' or 'RBApy' (default: 'SBML')
        :return: success
        :rtype: bool
        """
        assert export_format in {'SBML', 'RBApy'}, 'argument format must be "SBML" or "RBApy")'
        if export_format == 'SBML':
            return self.model.export(fname)
        elif export_format == 'RBApy':
            if fname.endswith('.xlsx'):
                return self.to_excel(fname)
            else:
                # export RBA model in RBA proprietary format
                if os.path.exists(fname) is False:
                    os.makedirs(fname)
                    print(f'RBA directory {fname} created')
                for component in components:
                    getattr(self, component).export_xml(fname)
                print(f'RBA model exported to: {fname}')
                return True



"""Implementation of BiocycGene class.

Peter Schubert, CCB, HHU Duesseldorf, November 2022
"""

import re
import xml.etree.ElementTree

from f2xba.utils.biocyc_utils import get_child_text, get_gene_products, get_sub_obj_ids


class BiocycGene:

    def __init__(self, biocyc_id):
        self.id = biocyc_id
        self.name = ''
        self.synonyms = ''
        self.locus = ''
        self.proteins = []
        self.rnas = []
        self.tus = []

    @staticmethod
    def get_genes(file_name):
        """Retrieve Gene data from biocyc export.

        """
        tree = xml.etree.ElementTree.parse(file_name)
        root = tree.getroot()

        data = {}
        for el in root.findall('Gene'):
            # we are only interested at genes with gene products defined
            if el.find('product') is None:
                continue
            biocyc_id = el.get('ID').split(':')[1]
            bc_gene = BiocycGene(biocyc_id)
            bc_gene.name = get_child_text(el, 'common-name')
            bc_gene.synonyms = re.sub(r'\|', ',', get_child_text(el, 'synonym'))
            bc_gene.locus = get_child_text(el, 'accession-1')
            bc_gene.proteins = get_gene_products(el, 'product', 'Protein')
            bc_gene.rnas = get_gene_products(el, 'product', 'RNA')
            bc_gene.tus = get_sub_obj_ids(el, 'component-of', 'Transcription-Unit')
            # bc_gene.dna_dir = get_child_text(el, 'transcription-direction')

            data[biocyc_id] = bc_gene
        return data



"""Implementation SbmlSpecies class.

based on XbaSpecies implementation

Peter Schubert, HHU Duesseldorf, February 2023
"""
import re

from .sbml_sbase import SbmlSBase


class SbmlSpecies(SbmlSBase):

    def __init__(self, s_species):
        """Instantiate SbmlSpecies instance with data from s_species.

        s_species is pandas Series with Series.name attribute set species id and
        several mandatory and optional attributes based on SBML specifications.

        - mandatory attributes:
            - Series.name: str - species id
            - 'compartment': str
            - 'constant': bool
            - 'boundaryCondition': bool
            - 'hasOnlySubstanceUnits': bool

        - optional attributes:
            - 'name': str - handled in parent class
            - 'sboterm': str - handled in parent class
            - 'metaid': str - handled in parent class
            - 'miriamAnnotation': str - handled in parent class
            - 'notes': str - handled in parent class
            - 'fbcCharge': float
            - 'fbcChemicalFormula': str

        From 'miriamAnnotation' references with Chebi and Seed ids are extracted.

        :param s_species: species data from SBML import
        :type s_species: pandas Series
        """
        super().__init__(s_species)
        self.compartment = s_species['compartment']
        self.constant = s_species.get('constant', False)
        self.boundary_condition = s_species.get('boundaryCondition', False)
        self.has_only_substance_units = s_species.get('hasOnlySubstanceUnits', False)
        if 'substanceUnits' in s_species:
            self.substance_units = s_species['substanceUnits']
        if 'fbcCharge' in s_species:
            self.charge = s_species['fbcCharge']
        if 'fbcChemicalFormula' in s_species:
            self.formula = s_species['fbcChemicalFormula']

    @property
    def chebi_refs(self):
        return [re.sub('CHEBI:', '', ref) for ref in self.miriam_annotation.get_qualified_refs('bqbiol:is', 'chebi')]

    @property
    def kegg_refs(self):
        return self.miriam_annotation.get_qualified_refs('bqbiol:is', 'kegg.compound')

    @property
    def seed_refs(self):
        return self.miriam_annotation.get_qualified_refs('bqbiol:is', 'seed.compound')

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str, int, float or bool
        """
        setattr(self, attribute, value)

    def to_dict(self):
        data = super().to_dict()
        data['compartment'] = self.compartment
        data['constant'] = self.constant
        data['boundaryCondition'] = self.boundary_condition
        data['hasOnlySubstanceUnits'] = self.has_only_substance_units

        if hasattr(self, 'substance_units'):
            data['substanceUnits'] = self.substance_units
        if hasattr(self, 'charge'):
            data['fbcCharge'] = self.charge
        if hasattr(self, 'formula'):
            data['fbcChemicalFormula'] = self.formula
        return data



"""Implementation of NcbiData class.

Extract nucleotide information from NCBI using E-utils EFetch.

NCBI's Disclaimer and Copyright notice
(https://www.ncbi.nlm.nih.gov/About/disclaimer.html).

also check: NCBI Entrez Programming Utilities Help

Peter Schubert, CCB, HHU Duesseldorf, January 2023
"""

from collections import defaultdict

from .ncbi_chromosome import NcbiChromosome


class NcbiData:

    def __init__(self, chromosome2accids, ncbi_dir):
        """Initialize

        Download NCBI nucleotide information for given accession ids.
        Use stored file, if found in ncbi_dir.

        :param chromosome2accids: Mapping chromosome id to GeneBank accession_id
        :type chromosome2accids: dict (key: chromosome id, str; value: Genbank accession_id, str)
        :param ncbi_dir: directory where ncbi exports are stored
        :type ncbi_dir: str
        """
        self.chromosomes = {}
        for chrom_id, accession_id in chromosome2accids.items():
            self.chromosomes[chrom_id] = NcbiChromosome(chrom_id, accession_id, ncbi_dir)

        # mapping of NCBI record loci to feature records and proteins across chromosomes
        self.locus2record = {}
        self.locus2protein = {}
        for chrom_id, chrom in self.chromosomes.items():
            self.locus2record.update(chrom.mrnas)
            self.locus2record.update(chrom.rrnas)
            self.locus2record.update(chrom.trnas)
            self.locus2protein.update(chrom.proteins)

        # mapping of gene product label to NCBI locus (including NCBI old_locus_tag)
        self.label2locus = {}
        self.update_label2locus()

    def update_label2locus(self):
        self.label2locus = {}
        for locus, record in self.locus2record.items():
            self.label2locus[locus] = locus
            if hasattr(record, 'old_locus') and record.old_locus is not None:
                self.label2locus[record.old_locus] = locus

    def modify_attributes(self, df_modify_attrs):
        """modify attribute values of ncbi feature records

        e.g. update 'locus' or 'old_locus' attributes to improve mapping with model loci

        :param df_modify_attrs: table with 'attribute', 'value' columns and index set to gene locus
        :type df_modify_attrs: pandas DataFrame
        """
        for locus, row in df_modify_attrs.iterrows():
            if locus in self.locus2record:
                record = self.locus2record[locus]
                record.modify_attribute(row['attribute'], row['value'])
            else:
                print(f'{locus} not found in NCBI data export')
        self.update_label2locus()

    def get_gc_content(self, chromosome_id=None):
        """Retrieve GC content accross all or a specifiec chromosome

        :param chromosome_id: specific chromosome id
        :type chromosome_id: str or None (optional: default: None)
        :return: GC content
        :rtype: float
        """
        if chromosome_id is not None:
            chrom_ids = [chromosome_id]
        else:
            chrom_ids = self.chromosomes.keys()

        total_nts = 0
        total_gc = 0
        for chrom_id in chrom_ids:
            chromosome = self.chromosomes[chrom_id]
            total_nts += sum(chromosome.nt_composition.values())
            total_gc += chromosome.nt_composition['G'] + chromosome.nt_composition['C']
        return total_gc / total_nts

    def get_mrna_avg_composition(self, chromosome_id=None):
        """Retrieve average mrna composition across all or a chromosome

        :param chromosome_id: specific chromosome id
        :type chromosome_id: str or None (optional: default: None)
        :return: relative mrna nucleotide composition
        :rtype: dict (key: nucleotide id, val: fequency/float)
        """
        chrom_ids = [chromosome_id] if chromosome_id is not None else self.chromosomes.keys()

        nt_comp = defaultdict(int)
        for chrom_id in chrom_ids:
            chrom = self.chromosomes[chrom_id]
            for locus, feature in chrom.mrnas.items():
                for nt, count in feature.spliced_nt_composition.items():
                    nt_comp[nt] += count
        total = sum(nt_comp.values())
        return {nt: count / total for nt, count in nt_comp.items()}
